# coding: utf-8

### WIKIPEDIA ########################################################################################
# Code for querying Wikipedia and parsing articles.
# The parser is as-is, it handles a lot but not everything.

# Author: Tom De Smedt.
# Copyright (c) 2007 by Tom De Smedt.
# See LICENSE.txt for details.

import re
from xml.dom import minidom
from urllib import quote

from url import URLAccumulator
from html import replace_entities, strip_tags
from cache import Cache
import mimetex

def clear_cache():
    Cache("wikipedia").clear()

### WIKIPEDIA PAGE MISSING ERROR #####################################################################

class WikipediaPageMissing(Exception):
    def __str__(self): return str(self.__class__)

### WIKIPEDIA LANGUAGES ##############################################################################

languages = {
    "aa"           : u"Afar",
    "ab"           : u"Abkhaz",
    "af"           : u"Afrikaans",
    "ak"           : u"Akan",
    "als"          : u"Alemannic",
    "am"           : u"Amharic",
    "an"           : u"Aragonese",
    "ang"          : u"Old English",
    "ar"           : u"Arabic",
    "arc"          : u"Aramaic",
    "as"           : u"Assamese",
    "ast"          : u"Asturian",
    "av"           : u"Avar",
    "ay"           : u"Aymara",
    "az"           : u"Azerbaijani",
    "ba"           : u"Bashkir",
    "bar"          : u"Bavarian",
    "bat-smg"      : u"Samogitian",
    "be"           : u"Belarusian",
    "bg"           : u"Bulgarian",
    "bh"           : u"Bihara",
    "bi"           : u"Bislama",
    "bm"           : u"Bambara",
    "bn"           : u"Bengali",
    "bo"           : u"Tibetan",
    "bpy"          : u"Bishnupriya Manipuri",
    "br"           : u"Breton",
    "bs"           : u"Bosnian",
    "bug"          : u"Buginese",
    "bxr"          : u"Buryat (Russia)",
    "ca"           : u"Catalan",
    "cbk-zam"      : u"Zamboanga Chavacano",
    "cdo"          : u"Min Dong",
    "ce"           : u"Chechen",
    "ceb"          : u"Cebuano",
    "ch"           : u"Chamorro",
    "cho"          : u"Choctaw",
    "chr"          : u"Cherokee",
    "chy"          : u"Cheyenne",
    "co"           : u"Corsican",
    "cr"           : u"Cree",
    "crh"          : u"Crimean Tatar",
    "crh-latn"     : u"Crimean Tatar (Latin)",
    "crh-cyrl"     : u"Crimean Tatar (Cyrillic)",
    "cs"           : u"Czech",
    "csb"          : u"Cassubian",
    "cu"           : u"Old Church Slavonic (ancient language)",
    "cv"           : u"Chuvash",
    "cy"           : u"Welsh",
    "da"           : u"Danish",
    "de"           : u"German",
    "diq"          : u"Zazaki",
    "dk"           : u"Danish",
    "dv"           : u"Dhivehi",
    "dz"           : u"Bhutani",
    "ee"           : u"Ewe",
    "el"           : u"Greek",
    "eml"          : u"Emilian-Romagnol/Sammarinese",
    "en"           : u"English",
    "eo"           : u"Esperanto",
    "es"           : u"Spanish",
    "et"           : u"Estonian",
    "eu"           : u"Basque",
    "fa"           : u"Persian",
    "ff"           : u"Fulah",
    "fi"           : u"Finnish",
    "fiu-vro"      : u"Voro",
    "fj"           : u"Fijian",
    "fo"           : u"Faroese",
    "fr"           : u"French",
    "frp"          : u"Franco-Provencal/Arpitan",
    "fur"          : u"Friulian",
    "fy"           : u"Frisian",
    "ga"           : u"Irish",
    "gd"           : u"Scots Gaelic",
    "gl"           : u"Gallegan",
    "glk"          : u"Gilaki",
    "gn"           : u"Guarani",
    "got"          : u"Gothic",
    "gsw"          : u"Alemannic",
    "gu"           : u"Gujarati",
    "gv"           : u"Manx",
    "ha"           : u"Hausa",
    "haw"          : u"Hawaiian",
    "he"           : u"Hebrew",
    "hi"           : u"Hindi",
    "hil"          : u"Hiligaynon",
    "ho"           : u"Hiri Motu",
    "hr"           : u"Croatian",
    "hsb"          : u"Upper Sorbian",
    "ht"           : u"Haitian",
    "hu"           : u"Hungarian",
    "hy"           : u"Armenian",
    "hz"           : u"Herero",
    "ia"           : u"Interlingua (IALA)",
    "id"           : u"Indonesian",
    "ie"           : u"Interlingue (Occidental)",
    "ig"           : u"Igbo",
    "ii"           : u"Sichuan Yi",
    "ik"           : u"Inupiak",
    "ilo"          : u"Ilokano",
    "io"           : u"Ido",
    "is"           : u"Icelandic",
    "it"           : u"Italian",
    "iu"           : u"Inuktitut",
    "ja"           : u"Japanese",
    "jbo"          : u"Lojban",
    "jv"           : u"Javanese",
    "ka"           : u"Georgian",
    "kaa"          : u"Karakalpak",
    "kab"          : u"Kabyle",
    "kg"           : u"KiKongo",
    "ki"           : u"Kikuyu",
    "kj"           : u"Kuanyama",
    "kk"           : u"Kazakh",
    "kk-cn"        : u"Kazakh Arabic",
    "kk-kz"        : u"Kazakh Cyrillic",
    "kk-tr"        : u"Kazakh Latin",
    "kl"           : u"Greenlandic",
    "km"           : u"Cambodian",
    "kn"           : u"Kannada",
    "ko"           : u"Korean",
    "kr"           : u"Kanuri",
    "ks"           : u"Kashmiri",
    "ksh"          : u"Ripuarian",
    "ku"           : u"Kurdish",
    "kv"           : u"Komi",
    "kw"           : u"Cornish",
    "ky"           : u"Kirghiz",
    "la"           : u"Latin",
    "lad"          : u"Ladino",
    "lbe"          : u"Lak",
    "lb"           : u"Luxemburguish",
    "lg"           : u"Ganda",
    "li"           : u"Limburgian",
    "lij"          : u"Ligurian",
    "lld"          : u"Ladin",
    "lmo"          : u"Lombard",
    "ln"           : u"Lingala",
    "lo"           : u"Laotian",
    "lt"           : u"Lithuanian",
    "lv"           : u"Latvian",
    "lzz"          : u"Laz",
    "map-bms"      : u"Banyumasan",
    "mg"           : u"Malagasy",
    "mh"           : u"Marshallese",
    "mi"           : u"Maori",
    "minnan"       : u"Min-nan",
    "mk"           : u"Macedonian",
    "ml"           : u"Malayalam",
    "mn"           : u"Mongoloian",
    "mo"           : u"Moldovan",
    "mr"           : u"Marathi",
    "ms"           : u"Malay",
    "mt"           : u"Maltese",
    "mus"          : u"Creek",
    "my"           : u"Burmese",
    "mzn"          : u"Mazandarin",
    "na"           : u"Nauruan",
    "nah"          : u"Nahuatl",
    "nan"          : u"Min-nan",
    "nap"          : u"Neapolitan",
    "nb"           : u"Norwegian (Bokmal)",
    "nds"          : u"Low German",
    "nds-nl"       : u"Dutch Low Saxon",
    "ne"           : u"Nepali",
    "new"          : u"Newar/Nepal Bhasa",
    "ng"           : u"Ndonga",
    "nl"           : u"Dutch",
    "nn"           : u"Norwegian (Nynorsk)",
    "no"           : u"Norwegian",
    "non"          : u"Old Norse",
    "nov"          : u"Novial",
    "nrm"          : u"Norman",
    "nv"           : u"Navajo",
    "ny"           : u"Chichewa",
    "oc"           : u"Occitan",
    "om"           : u"Oromo",
    "or"           : u"Oriya",
    "os"           : u"Ossetic",
    "pa"           : u"Punjabi",
    "pag"          : u"Pangasinan",
    "pam"          : u"Pampanga",
    "pap"          : u"Papiamentu",
    "pdc"          : u"Pennsylvania German",
    "pih"          : u"Norfuk/Pitcairn/Norfolk",
    "pi"           : u"Pali",
    "pl"           : u"Polish",
    "pms"          : u"Piedmontese",
    "ps"           : u"Pashto",
    "pt"           : u"Portuguese",
    "pt-br"        : u"Brazilian Portuguese",
    "qu"           : u"Quechua",
    "rm"           : u"Raeto-Romance",
    "rmy"          : u"Vlax Romany",
    "rn"           : u"Kirundi",
    "ro"           : u"Romanian",
    "roa-rup"      : u"Aromanian",
    "roa-tara"     : u"Tarantino",
    "ru"           : u"Russian",
    "ru-sib"       : u"Siberian/North Russian",
    "rw"           : u"Kinyarwanda",
    "sa"           : u"Sanskrit",
    "sc"           : u"Sardinian",
    "scn"          : u"Sicilian",
    "sco"          : u"Scots",
    "sd"           : u"Sindhi",
    "se"           : u"Northern Sami",
    "sg"           : u"Sango",
    "sh"           : u"Serbocroatian",
    "si"           : u"Sinhalese",
    "simple"       : u"Simple English",
    "sk"           : u"Slovak",
    "sl"           : u"Slovenian",
    "sm"           : u"Samoan",
    "sn"           : u"Shona",
    "so"           : u"Somali",
    "sq"           : u"Albanian",
    "sr"           : u"Serbian",
    "sr-ec"        : u"Serbian cyrillic ekavian",
    "sr-jc"        : u"Serbian cyrillic iyekvian",
    "sr-el"        : u"Serbian latin ekavian",
    "sr-jl"        : u"Serbian latin iyekavian",
    "ss"           : u"Swati",
    "st"           : u"Southern Sotho",
    "su"           : u"Sundanese",
    "sv"           : u"Swedish",
    "sw"           : u"Swahili",
    "ta"           : u"Tamil",
    "te"           : u"Telugu",
    "tet"          : u"Tetun",
    "tg"           : u"Tajik",
    "th"           : u"Thai",
    "ti"           : u"Tigrinya",
    "tk"           : u"Turkmen",
    "tl"           : u"Tagalog (Filipino)",
    "tlh"          : u"Klingon",
    "tn"           : u"Setswana",
    "to"           : u"Tonga (Tonga Islands)",
    "tokipona"     : u"Toki Pona",
    "tp"           : u"Toki Pona",
    "tpi"          : u"Tok Pisin",
    "tr"           : u"Turkish",
    "ts"           : u"Tsonga",
    "tt"           : u"Tatar",
    "tum"          : u"Tumbuka",
    "tw"           : u"Twi",
    "ty"           : u"Tahitian",
    "tyv"          : u"Tyvan",
    "udm"          : u"Udmurt",
    "ug"           : u"Uyghur",
    "uk"           : u"Ukrainian",
    "ur"           : u"Urdu",
    "uz"           : u"Uzbek",
    "ve"           : u"Venda",
    "vec"          : u"Venetian",
    "vi"           : u"Vietnamese",
    "vls"          : u"West Flemish",
    "vo"           : u"Volapuk",
    "wa"           : u"Walloon",
    "war"          : u"Waray-Waray",
    "wo"           : u"Wolof",
    "wuu"          : u"Wu",
    "xal"          : u"Kalmyk",
    "xh"           : u"Xhosan",
    "yi"           : u"Yiddish",
    "yo"           : u"Yoruba",
    "za"           : u"Zhuang",
    "zea"          : u"Zealandic",
    "zh"           : u"Chinese", # correct?
    "zh-cfr"       : u"Min-nan",
    "zh-classical" : u"Classical Chinese/Literary Chinese",
    "zh-cn"        : u"Simplified",
    "zh-hk"        : u"Traditional (Hong Kong)",
    "zh-min-nan"   : u"Min-nan",
    "zh-sg"        : u"Simplified (Singapore)",
    "zh-tw"        : u"Traditional",
    "zh-yue"       : u"Cantonese",
    "zu"           : u"Zulu",
}

### WIKIPEDIALINK ####################################################################################
# Currently not in use.

class WikipediaLink:
    
    def __init__(self, page, anchor=u"", display=u""):
        
        self.page = page
        self.anchor = anchor
        self.display = display
        
    def __str__(self):
        return self.page.encode("utf-8")
    
    def __unicode__(self):
        return self.page
        
### WIKIPEDIAPARAGRAPH ###############################################################################

class WikipediaParagraph(list):
    
    def __init__(self, title=u"", main=[], related=[], tables=[]):
        
        self.title = title
        self.main = main
        self.related = related
        self.tables = []
        
        self.depth = 0
        self.parent = None
        self.children = []
        
    def __str__(self):
        s = "\n\n".join(self)
        return s.encode("utf-8")

    def __unicode__(self):
        s = "\n\n".join(self)
        return s
 
### WIKIPEDIAIMAGE ###################################################################################
    
class WikipediaImage:
    
    def __init__(self, path, description=u"", links=[], properties=[]):
        
        self.path = path
        self.description = description
        self.links = links
        self.properties = properties
        
    def __str__(self):
        return self.path.encode("utf-8")

    def __unicode__(self):
        return self.path

### WIKIPEDIAREFERENCES ##############################################################################

class WikipediaReference:
    
    def __init__(self, title=u"", url=u""):
        
        self.title = title
        self.url = url
        self.author    = u""
        self.first     = u""
        self.last      = u""
        self.journal   = u""
        self.publisher = u""
        self.date      = u""
        self.year      = u""
        self.id        = u""
        
        self.note      = u""
        
    def __str__(self):
        
        s = ""
        for key in ["note", "author", "title", "journal", "publisher", "date", "id", "url"]:
            value = getattr(self, key)
            if value != "":
                s += value.rstrip(".,") + ", "
        
        s = s.strip(", \n")
        return s.encode("utf-8")
        
    def __unicode__(self):
        return str(self).decode("utf-8")

### WIKIPEDIATABLE ###################################################################################

class WikipediaTable(list):
    
    def __init__(self, title=u"", properties=u"", paragraph=None):
        
        self.title = u""
        self.properties = properties
        self.paragraph = None
      
class WikipediaTableRow(list):
    
    def __init__(self, heading=False, properties=u""):
        
        self.properties = properties

class WikipediaTableCell(unicode):
    
    def __init__(self, data):
        
        unicode.__init__(self, data)
        self.properties = u""

### WIKIPEDIAPAGE ####################################################################################

class WikipediaPage:
    
    def __init__(self, title, markup, light=False, full_strip=True):
        
        """ Wikipedia page parser.
        
        The expected markup is the stuff in Wikipedia's edit textarea.
        With light=True, it will onlt parse links to other articles (which is faster).
        With full_strip=False, it will preserve some HTML markup (links, bold, italic).
        
        """
        
        self.title = title
        self.markup = markup
        self.full_strip = full_strip
        
        self.disambiguation = []
        self.categories = []
        self.links = []
        
        self.paragraphs = []
        self.images = []
        self.tables = []
        self.references = []
        self.translations = {}
        self.important = []
        
        # Main regular expressions used in the parser.
        self.re = {
            "disambiguation" : r"\{\{dablink\|(.*)\}\}",
            "category"       : r"\[\[[:]{0,1}Category:(.*?)\]\]",
            "link"           : r"\[\[([^\:]*?)\]\]",
            "image"          : re.compile(r"\[\[Image:[^\[]*\|.*\]\]", re.I),
            "gallery"        : re.compile("<gallery>(.*?)</gallery>", re.DOTALL),
            "table"          : re.compile(r"\{\|.*?\|\}", re.DOTALL),
            "html-table"     : re.compile(r"<table.*?>.*?</table>", re.DOTALL),
            "reference"      : re.compile(r"<ref.*?>.*?</ref>", re.DOTALL),
            "citation"       : re.compile(r"\{\{cite.*?\}\}", re.DOTALL),
            "url"            : r"\[(http\://.*?)\]",            
            "preformatted"   : re.compile(r"<pre.*?>.*?</pre>", re.DOTALL),
            "translation"    : r"\[\[([^\].]*?):(.*?)\]\]",
            "bold"           : r"\'\'\'(.*?)\'\'\'",
            "comment"        : re.compile(r"<!--.*?-->", re.DOTALL),       
        }
        
        # In the process of stripping references and citations from the markup,
        # they are temporarily marked by this pattern.
        # Don't use any regex characters in it.
        self.ref = "--REF--"
        
        self.parse(light)

    def __unicode__(self):
        
        str = u""
        for paragraph in self.paragraphs:
            str += paragraph.title+"\n\n"
            for textblock in paragraph:
                str += unicode(textblock)+"\n\n"
        
        return str
        
    def __str__(self):
        s = ""
        for p in self.paragraphs:
            s += (p.title.encode("utf-8") + "\n\n").lstrip("\n")
            s += (str(p) + "\n\n").lstrip("\n")
        return s
    
    def __unicode__(self):
        return str(self).decode("utf-8")

    def parse(self, light=False):

        """ Parses data from Wikipedia page markup.

        The markup comes from Wikipedia's edit page.
        We parse it here into objects containing plain text.
        The light version parses only links to other articles, it's faster than a full parse.    
        
        """

        markup = self.markup
        
        self.disambiguation = self.parse_disambiguation(markup)
        self.categories = self.parse_categories(markup)
        self.links = self.parse_links(markup)
        
        if not light:
        
            # Conversion of HTML markup to Wikipedia markup.
            markup = self.convert_pre(markup)
            markup = self.convert_li(markup)
            markup = self.convert_table(markup)
            markup = replace_entities(markup)
        
            # Harvest references from the markup
            # and replace them by footnotes.
            markup = markup.replace("{{Cite", "{{cite")
            markup = re.sub("\{\{ {1,2}cite", "{{cite", markup)
            self.references, markup = self.parse_references(markup)

            # Make sure there are no legend linebreaks in image links.
            # Then harvest images and strip them from the markup.
            markup = re.sub("\n+(\{\{legend)", "\\1", markup)
            self.images, markup = self.parse_images(markup)
            self.images.extend(self.parse_gallery_images(markup))
            
            self.paragraphs = self.parse_paragraphs(markup)
            self.tables = self.parse_tables(markup)
            self.translations = self.parse_translations(markup)
            self.important = self.parse_important(markup)
    
    def plain(self, markup):
        
        """ Strips Wikipedia markup from given text.
        
        This creates a "plain" version of the markup,
        stripping images and references and the like.
        Does some commonsense maintenance as well,
        like collapsing multiple spaces.
        If you specified full_strip=False for WikipediaPage instance,
        some markup is preserved as HTML (links, bold, italic).
        
        """
        
        # Strip bold and italic.
        if self.full_strip:
            markup = markup.replace("'''", "")
            markup = markup.replace("''", "")
        else:
            markup = re.sub("'''([^']*?)'''", "<b>\\1</b>", markup)
            markup = re.sub("''([^']*?)''", "<i>\\1</i>", markup)
        
        # Strip image gallery sections.
        markup = re.sub(self.re["gallery"], "", markup)
        
        # Strip tables.
        markup = re.sub(self.re["table"], "", markup)
        markup = markup.replace("||", "")
        markup = markup.replace("|}", "")
        
        # Strip links, keeping the display alias.
        # We'll strip the ending ]] later.
        if self.full_strip:
            markup = re.sub(r"\[\[[^\]]*?\|", "", markup)
        else:
            markup = re.sub(r"\[\[([^]|]*|)\]\]", '<a href="\\1">\\1</a>', markup)
            markup = re.sub(r"\[\[([^]|]*|)\|([^]]*)\]\]", '<a href="\\1">\\2</a>', markup)    

        # Strip translations, users, etc.
        markup = re.sub(self.re["translation"], "", markup)
        
        # This math TeX is not supported:
        markup = markup.replace("\displaytyle", "")
        markup = markup.replace("\textstyle", "")
        markup = markup.replace("\scriptstyle", "")
        markup = markup.replace("\scriptscriptstyle", "")
        
        # Before stripping [ and ] brackets,
        # make sure they are retained inside <math></math> equations.
        markup = re.sub("(<math>.*?)\[(.*?</math>)", "\\1MATH___OPEN\\2", markup)
        markup = re.sub("(<math>.*?)\](.*?</math>)", "\\1MATH___CLOSE\\2", markup)
        markup = markup.replace("[", "")
        markup = markup.replace("]", "")
        markup = markup.replace("MATH___OPEN", "[")
        markup = markup.replace("MATH___CLOSE", "]")
        
        # a) Strip references.
        # b) Strip <ref></ref> tags.
        # c) Strip <ref name="" /> tags.
        # d) Replace --REF--(12) by [12].
        # e) Remove space between [12] and trailing punctuation .,
        # f) Remove HTML comment <!-- -->
        # g) Keep the Latin Extended-B template: {{latinx| }}
        # h) Strip Middle-Earth references.
        # i) Keep quotes: {{quote| }}
        # j) Remove templates
        markup = re.sub(self.re["reference"], "", markup)                  # a
        markup = re.sub("</{0,1}ref.*?>", "", markup)                      # b
        markup = re.sub("<ref name=\".*?\" {0,1}/>", "", markup)           # c
        markup = re.sub(self.ref+"\(([0-9]*?)\)", "[\\1] ", markup)        # d
        markup = re.sub("\] ([,.\"\?\)])", "]\\1", markup)                 # e
        markup = re.sub(self.re["comment"], "", markup)                    # f
        markup = re.sub("\{\{latinx\|(.*?)\}\}", "\\1", markup)            # g
        markup = re.sub("\{\{ME-ref.*?\}\}", "", markup)                   # h
        markup = re.sub("\{\{quote\|(.*?)\}\}", "\"\\1\"", markup)         # i
        markup = re.sub(re.compile("\{\{.*?\}\}", re.DOTALL), "", markup)  # j
        markup = markup.replace("}}", "")
        
        # Collapse multiple spaces between words,
        # unless they appear in preformatted text.
        markup = re.sub("<br.*?/{0,1}>", " ", markup)
        markup = markup.split("\n")
        for i in range(len(markup)):
            if not markup[i].startswith(" "):
                markup[i] = re.sub(r"[ ]+", " ", markup[i])
        markup = "\n".join(markup)
        markup = markup.replace(" .", ".")
        
        # Strip all HTML except <math> tags.
        if self.full_strip:
            markup = strip_tags(markup, exclude=["math"], linebreaks=True)
        
        markup = markup.strip()
        return markup
    
    def convert_pre(self, markup):
        
        """ Substitutes <pre> to Wikipedia markup by adding a space at the start of a line.
        """
        
        for m in re.findall(self.re["preformatted"], markup):
            markup = markup.replace(m, m.replace("\n", "\n "))
            markup = re.sub("<pre.*?>\n{0,}", "", markup)
            markup = re.sub("\W{0,}</pre>", "", markup)
        
        return markup
    
    def convert_li(self, markup):

        """ Subtitutes <li> content to Wikipedia markup.
        """
        
        for li in re.findall("<li;*?>", markup):
            markup = re.sub(li, "\n* ", markup)
        markup = markup.replace("</li>", "")
            
        return markup
    
    def convert_table(self, markup):
        
        """ Subtitutes <table> content to Wikipedia markup.
        """
        
        for table in re.findall(self.re["html-table"], markup):
            wiki = table
            wiki = re.sub(r"<table(.*?)>", "{|\\1", wiki)
            wiki = re.sub(r"<tr(.*?)>", "|-\\1", wiki)
            wiki = re.sub(r"<td(.*?)>", "|\\1|", wiki)
            wiki = wiki.replace("</td>", "\n")
            wiki = wiki.replace("</tr>", "\n")
            wiki = wiki.replace("</table>", "\n|}")
            markup = markup.replace(table, wiki)
        
        return markup
    
    def parse_links(self, markup):
        
        """ Returns a list of internal Wikipedia links in the markup.

        # A Wikipedia link looks like:
        # [[List of operating systems#Embedded | List of embedded operating systems]]
        # It does not contain a colon, this indicates images, users, languages, etc.
        
        The return value is a list containing the first part of the link,
        without the anchor.

        """
        
        links = []
        m = re.findall(self.re["link"], markup)
        for link in m:
            # We don't like [[{{{1|Universe (disambiguation)}}}]]
            if link.find("{") >= 0:
                link = re.sub("\{{1,3}[0-9]{0,2}\|", "", link)
                link = link.replace("{", "")
                link = link.replace("}", "")            
            link = link.split("|")
            link[0] = link[0].split("#")
            page = link[0][0].strip()
            #anchor = u""
            #display = u""
            #if len(link[0]) > 1: 
            #    anchor = link[0][1].strip()
            #if len(link) > 1: 
            #    display = link[1].strip()
            if not page in links:
                links.append(page)
                #links[page] = WikipediaLink(page, anchor, display)
        
        links.sort()
        return links

    def parse_images(self, markup, treshold=6):
        
        """ Returns a list of images found in the markup.
        
        An image has a pathname, a description in plain text
        and a list of properties Wikipedia uses to size and place images.

        # A Wikipedia image looks like:
        # [[Image:Columbia Supercomputer - NASA Advanced Supercomputing Facility.jpg|right|thumb|
        #   The [[NASA]] [[Columbia (supercomputer)|Columbia Supercomputer]].]]
        # Parts are separated by "|".
        # The first part is the image file, the last part can be a description.
        # In between are display properties, like "right" or "thumb".
        
        """
        
        images = []
        m = re.findall(self.re["image"], markup)
        for p in m:
            p = self.parse_balanced_image(p)
            img = p.split("|")
            path = img[0].replace("[[Image:", "").strip()
            description = u""
            links = {}
            properties = []
            if len(img) > 1:
                img = "|".join(img[1:])
                links = self.parse_links(img)
                properties = self.plain(img).split("|")
                description = u""
                # Best guess: an image description is normally
                # longer than six characters, properties like
                # "thumb" and "right" are less than six characters.
                if len(properties[-1]) > treshold:
                    description = properties[-1]
                    properties = properties[:-1]
            img = WikipediaImage(path, description, links, properties)
            images.append(img)
            markup = markup.replace(p, "")
        
        return images, markup.strip()
    
    def parse_balanced_image(self, markup):
        
        """ Corrects Wikipedia image markup.

        Images have a description inside their link markup that 
        can contain link markup itself, make sure the outer "[" and "]" brackets 
        delimiting the image are balanced correctly (e.g. no [[ ]] ]]).

        Called from parse_images().

        """

        opened = 0
        closed = 0
        for i in range(len(markup)):
            if markup[i] == "[": opened += 1
            if markup[i] == "]": closed += 1
            if opened == closed:
                return markup[:i+1]
                
        return markup

    def parse_gallery_images(self, markup):
        
        """ Parses images from the <gallery></gallery> section.
        
        Images inside <gallery> tags do not have outer "[[" brackets.
        Add these and then parse again.
        
        """
        
        gallery = re.search(self.re["gallery"], markup)
        if gallery:
            gallery = gallery.group(1)
            gallery = gallery.replace("Image:", "[[Image:")
            gallery = gallery.replace("\n", "]]\n")
            images, markup = self.parse_images(gallery)
            return images
        
        return []
    
    def parse_paragraph(self, markup):
        
        """ Creates a list from lines of text in a paragraph.
        
        Each line of text is a new item in the list,
        except lists and preformatted chunks (<li> and <pre>),
        these are kept together as a single chunk.
        
        Lists are formatted using parse_paragraph_list().
        
        Empty lines are stripped from the output.
        Indentation (i.e. lines starting with ":") is ignored.
        
        Called from parse_paragraphs() method.
        
        """
        
        s = self.plain(markup)
        # Add an extra linebreak between the last list item
        # and the normal line following after it, so they don't stick together, e.g.
        # **[[Alin Magic]], magic used in the videogame ''[[Rise of Nations: Rise of Legends]]''
        # In '''popular culture''':
        # * [[Magic (film)|''Magic'' (film)]], a 1978 film starring Anthony Hopkins and Ann-Margret
        s = re.sub(re.compile("\n([*#;].*?)\n([^*#?])", re.DOTALL), "\n\\1\n\n\\2", s)
        # This keeps list items with linebreaks 
        # between them nicely together.
        s = re.sub("\n{2,3}([*#;])", "\n\\1", s)
        chunks = []
        ch = ""
        i = 1
        for chunk in s.split("\n"):
            if chunk.startswith(":"):
                chunk = chunk.lstrip(":")
            if len(chunk.strip()) > 1:
                # Leave out taxoboxes and infoboxes.
                if not chunk.startswith("|"):
                    ch += chunk + "\n"
            if ch.strip() != "":
                if not re.search("^[ *#;]", chunk):
                    ch = self.parse_paragraph_list(ch)
                    chunks.append(ch.rstrip())
                    ch = ""

        if ch.strip() != "":
            ch = self.parse_paragraph_list(ch)
            chunks.append(ch.strip())
            
        return chunks        
    
    def parse_paragraph_list(self, markup, indent="\t"):
        
        """ Formats bullets and numbering of Wikipedia lists.
        
        List items are marked by "*", "#" or ";" at the start of a line.
        We treat ";" the same as "*",
        and replace "#" with real numbering (e.g. "2.").
        Sublists (e.g. *** and ###) get indented by tabs.
        
        Called from parse_paragraphs() method.
        
        """

        def lastleft(ch, str):
            n = 0
            while n < len(str) and str[n] == ch: n += 1
            return n        

        tally = [1 for i in range(10)]
        chunks = markup.split("\n")
        for i in range(len(chunks)):
            if chunks[i].startswith("#"):
                j = min(lastleft("#", chunks[i]), len(tally)-1)
                chunks[i] = indent*(j-1) + str(tally[j])+". " + chunks[i][j:]
                chunks[i] = chunks[i].replace(".  ", ". ")
                tally[j] += 1
                # Reset the numbering of sublists.
                for k in range(j+1, len(tally)): 
                    tally[k] = 1
            if chunks[i].startswith(";"):
                chunks[i] = "*" + chunks[i][1:]
            if chunks[i].startswith("*"):
                j = lastleft("*", chunks[i])  
                chunks[i] = indent*(j-1) + "* " + chunks[i][j:]
                chunks[i] = chunks[i].replace("*  ", "* ")
        
        return "\n".join(chunks)
    
    def parse_paragraph_heading_depth(self, markup):
        
        """ Returns the depth of a heading.
        
        The depth determines parent and child relations,
        which headings (and hence which paragraphs) are a child to a heading higher up.
        Returns 0 for <h1> =, 1 for <h2> ==, etc.
        
        Called from parse_paragraphs() method.
        
        """
        
        return markup.count("=")/2 - 1
        
    def connect_paragraph(self, paragraph, paragraphs):
        
        """ Create parent/child links to other paragraphs.
        
        The paragraphs parameters is a list of all the paragraphs
        parsed up till now.
        
        The parent is the previous paragraph whose depth is less.
        The parent's children include this paragraph.
        
        Called from parse_paragraphs() method.
        
        """

        if paragraph.depth > 0:
            n = range(len(paragraphs))
            n.reverse()
            for i in n:
                if paragraphs[i].depth == paragraph.depth-1:
                    paragraph.parent = paragraphs[i]
                    paragraphs[i].children.append(paragraph)
                    break
                    
        return paragraph          

    def parse_paragraph_references(self, markup):
        
        """ Updates references with content from specific paragraphs.
        
        The "references", "notes", "external links" paragraphs 
        are double-checked for references. Not all items in the list
        might have been referenced inside the article, or the item
        might contain more info than we initially parsed from it.
        
        Called from parse_paragraphs() method.
        
        """
        
        for chunk in markup.split("\n"):
            # We already parsed this, it contains the self.ref mark.
            # See if we can strip more notes from it.
            m = re.search(self.ref+"\(([0-9]*?)\)", chunk)
            if m:
                chunk = chunk.strip("* ")
                chunk = chunk.replace(m.group(0), "")
                chunk = self.plain(chunk)
                i = int(m.group(1))
                if chunk != "":
                    self.references[i-1].note = chunk
            # If it's not a citation we don't have this reference yet.
            elif chunk.strip().startswith("*") \
             and chunk.find("{{cite") < 0:
                chunk = chunk.strip("* ")
                chunk = self.plain(chunk)
                if chunk != "":
                    r = WikipediaReference()
                    r.note = chunk
                    self.references.append(r)
    
    def parse_paragraphs(self, markup):
        
        """ Returns a list of paragraphs in the markup.
        
        A paragraph has a title and multiple lines of plain text.
        A paragraph might have parent and child paragraphs,
        denoting subtitles or bigger chapters.
        
        A paragraph might have links to additional articles.
        
        Formats numbered lists by replacing # by 1.
        Formats bulleted sublists like ** or *** with indentation.
        
        """
        
        # Paragraphs to exclude.
        refs = ["references", "notes", "notes and references", "external links", "further reading"]
        exclude = ["see also", "media", "gallery", "related topics", "lists", "gallery", "images"]
        exclude.extend(refs)
        
        paragraphs = []
        paragraph = WikipediaParagraph(self.title)
        paragraph_data = ""
        for chunk in markup.split("\n"):
            
            # Strip each line of whitespace, 
            # unless it's a preformatted line (starts with a space).
            if not chunk.startswith(" "):
                chunk = chunk.strip()
                
            # A title wrapped in "=", "==", "==="...
            # denotes a new paragraphs section.
            if chunk.startswith("="):

                if paragraph.title.lower() in refs \
                or (paragraph.parent and paragraph.parent.title.lower() in refs):
                    self.parse_paragraph_references(paragraph_data)
                paragraph.extend(self.parse_paragraph(paragraph_data))
                paragraphs.append(paragraph)
                
                # Initialise a new paragraph.
                # Create parent/child links to other paragraphs.
                title = chunk.strip().strip("=")
                title = self.plain(title)
                paragraph = WikipediaParagraph(title)
                paragraph.depth = self.parse_paragraph_heading_depth(chunk)
                if paragraph.title.lower() not in exclude:
                    paragraph = self.connect_paragraph(paragraph, paragraphs)
                paragraph_data = ""
            
            # Underneath a title might be links to in-depth articles,
            # e.g. Main articles: Computer program and Computer programming
            # which in wiki markup would be {{main|Computer program|Computer programming}}
            # The second line corrects" {{Main|Credit (finance)}} or {{Main|Usury}}".
            elif re.search(re.compile("^{{main", re.I), chunk):
                paragraph.main = [link.strip("} ") for link in chunk.split("|")[1:]]
                paragraph.main = [re.sub(re.compile("}}.*?{{main", re.I), "", link) 
                                  for link in paragraph.main]
                
            # At the bottom might be links to related articles,
            # e.g. See also: Abundance of the chemical elements
            # which in wiki markup would be {{see also|Abundance of the chemical elements}}
            elif re.search(re.compile("^{{see {0,1}also", re.I), chunk):
                paragraph.related = [link.strip("} ") for link in chunk.split("|")[1:]]
                
            # Accumulate the data in this paragraph,
            # we'll process it once a new paragraph starts.
            else:
                paragraph_data += chunk +"\n"
                
        # Append the last paragraph.
        if paragraph.title.lower() in refs \
        or (paragraph.parent and paragraph.parent.title.lower() in refs):
            self.parse_paragraph_references(paragraph_data)
        paragraph.extend(self.parse_paragraph(paragraph_data))
        paragraphs.append(paragraph)

        # The "See also" paragraph is an enumeration of links
        # which we already parsed so don't show them.
        # We also did references, and other paragraphs are not that relevant.        
        paragraphs_exclude = []
        for paragraph in paragraphs:
            if paragraph.title.lower() not in exclude \
            and not (paragraph.parent and paragraph.parent.title.lower() in exclude):
                paragraphs_exclude.append(paragraph)
        
        if len(paragraphs_exclude) == 1 and \
           len(paragraphs_exclude[0]) == 0:
            return []
        
        return paragraphs_exclude    

    def parse_table_row(self, markup, row):

        """ Parses a row of cells in a Wikipedia table.
        
        Cells in the row are separated by "||".
        A "!" indicates a row of heading columns.
        Each cell can contain properties before a "|",
        # e.g. align="right" | Cell 2 (right aligned).       
        
        """
        
        if row == None:
            row = WikipediaTableRow()
           
        markup = markup.replace("!!", "||")
        for cell in markup.lstrip("|!").split("||"):
            # The "|" after the properties can't be part of a link.
            i = cell.find("|")
            j = cell.find("[[")
            if i>0 and (j<0 or i<j):
                data = self.plain(cell[i+1:])
                properties = cell[:i].strip()
            else:
                data = self.plain(cell)
                properties = u""
            cell = WikipediaTableCell(data)
            cell.properties = properties
            row.append(cell)
        
        return row

    def connect_table(self, table, chunk, markup):

        """ Creates a link from the table to paragraph and vice versa.
        
        Finds the first heading above the table in the markup.
        This is the title of the paragraph the table belongs to.
        
        """

        k = markup.find(chunk)
        i = markup.rfind("\n=", 0, k)
        j = markup.find("\n", i+1)
        paragraph_title = markup[i:j].strip().strip("= ")
        for paragraph in self.paragraphs:
            if paragraph.title == paragraph_title:
                paragraph.tables.append(table)
                table.paragraph = paragraph

    def parse_tables(self, markup):
        
        """ Returns a list of tables in the markup.

        A Wikipedia table looks like:
        {| border="1"
        |-
        |Cell 1 (no modifier - not aligned)
        |-
        |align="right" |Cell 2 (right aligned)
        |-
        |}

        """

        tables = []
        m = re.findall(self.re["table"], markup)
        for chunk in m:

            table = WikipediaTable()
            table.properties = chunk.split("\n")[0].strip("{|").strip()
            self.connect_table(table, chunk, markup)
                  
            # Tables start with "{|".
            # On the same line can be properties, e.g. {| border="1"
            # The table heading starts with "|+".
            # A new row in the table starts with "|-".
            # The end of the table is marked with "|}".            
            row = None
            for chunk in chunk.split("\n"):
                chunk = chunk.strip()
                if chunk.startswith("|+"):
                    title = self.plain(chunk.strip("|+"))
                    table.title = title
                elif chunk.startswith("|-"):
                    if row: 
                        row.properties = chunk.strip("|-").strip()
                        table.append(row)
                    row = None
                elif chunk.startswith("|}"):
                    pass
                elif chunk.startswith("|") \
                  or chunk.startswith("!"):
                    row = self.parse_table_row(chunk, row)
                        
            # Append the last row.
            if row: table.append(row)
            if len(table) > 0:
                tables.append(table)
        
        return tables

    def parse_references(self, markup):

        """ Returns a list of references found in the markup.
        
        References appear inline as <ref> footnotes, 
        http:// external links, or {{cite}} citations.
        We replace it with (1)-style footnotes.
        Additional references data is gathered in
        parse_paragraph_references() when we parse paragraphs.
        
        References can also appear in image descriptions,
        tables and taxoboxes, so they might not always pop up in a paragraph.
        
        The plain() method finally replaces (1) by [1].
        
        """
    
        references = []
        
        # A Wikipedia reference note looks like:
        # <ref>In 1946, [[ENIAC]] consumed an estimated 174 kW. 
        # By comparison, a typical personal computer may use around 400 W; 
        # over four hundred times less. {{Ref harvard|kempf1961|Kempf 1961|a}}</ref>
        m = re.findall(self.re["reference"], markup)
        for reference in m:
            reference = re.sub("<ref> {0,1}cite", "<ref>{{cite", reference)
            if not reference.strip().startswith("[http://") and \
               not re.search("\{\{cite", reference):
                r = WikipediaReference()
                r.note = self.plain(re.sub("</{0,1}ref.*?>", "", reference))
                if r.note != "":
                    references.append(r)
                    p = " "+self.ref+"("+str(len(references))+")"
                    markup = markup.replace(reference, p, 1)
            else:
                # References containing a citation or url 
                # are better handled by the next patterns.
                pass
        
        # A Wikipedia citation looks like:
        # {{cite journal
        # | last = Einstein 
        # | first = Albert
        # | authorlink = Albert Einstein
        # | title = Sidelights on Relativity (Geometry and Experience) 
        # | publisher = P. Dutton., Co 
        # | date = 1923}}
        m = re.findall(self.re["citation"], markup)
        for citation in m:
            c = citation.replace("\n", "")
            r = WikipediaReference()
            for key in r.__dict__.keys():
                value = re.search("\| {0,1}"+key+"(.*?)[\|}]", c)
                if value:
                    value = value.group(1)
                    value = value.replace("link", "")
                    value = value.strip().strip(" =[]")
                    value = self.plain(value)
                    setattr(r, key, value)
            if r.first != "" and r.last != "":
                r.author = r.first + " " + r.last
            references.append(r)
            p = " "+self.ref+"("+str(len(references))+")"
            markup = markup.replace(citation, p, 1)
        
        # A Wikipedia embedded url looks like:
        # [http://www.pbs.org/wnet/hawking/html/home.html ''Stephen Hawking's Universe'']
        m = re.findall(self.re["url"], markup)
        for url in m:
            r = WikipediaReference()
            i = url.find(" ")
            if i > 0:
                r.url = url[:i].strip()
                r.note = self.plain(url[i:])
            else:
                r.url = url.strip()
            references.append(r)
            p = r.note+" "+self.ref+"("+str(len(references))+")"
            markup = markup.replace("["+url+"]", p, 1)

        # Since we parsed all citations first and then all notes and urls,
        # the ordering will not be correct in the markup,
        # e.g. (1) (11) (12) (2) (3).
        sorted = []
        m = re.findall(self.ref+"\(([0-9]*)\)", markup)
        for i in m:
            sorted.append(references[int(i)-1])
            markup = markup.replace(
                self.ref+"("+i+")", 
                self.ref+"**("+str(len(sorted))+")"
                )
        markup = markup.replace(self.ref+"**", self.ref)
        for r in references:
            if r not in sorted:
                sorted.append(r)
        references = sorted

        return references, markup.strip()
                
    def parse_categories(self, markup):
        
        """ Returns a list of categories the page belongs to.

        # A Wikipedia category link looks like:
        # [[Category:Computing]]
        # This indicates the page is included in the given category.
        # If "Category" is preceded by ":" this indicates a link to a category.
        
        """
        
        categories = []
        m = re.findall(self.re["category"], markup)
        for category in m:
            category = category.split("|")
            page = category[0].strip()
            display = u""
            if len(category) > 1: 
                display = category[1].strip()
            #if not categories.has_key(page):
            #    categories[page] = WikipediaLink(page, u"", display)
            if not page in categories:
                categories.append(page)
                
        return categories
    
    def parse_translations(self, markup):
        
        """ Returns a dictionary of translations for the page title.
        
        A Wikipedia language link looks like: [[af:Rekenaar]].
        The parser will also fetch links like "user:" and "media:"
        but these are stripped against the dictionary of
        Wikipedia languages.
        
        You can get a translated page by searching Wikipedia
        with the appropriate language code and supplying
        the translated title as query.
        
        """
        
        global languages
        translations = {}
        m = re.findall(self.re["translation"], markup)
        for language, translation in m:
            if language in languages:
                translations[language] = translation
         
        return translations
    
    def parse_disambiguation(self, markup):
        
        """ Gets the Wikipedia disambiguation page for this article.
        
        A Wikipedia disambiguation link refers to other pages
        with the same title but of smaller significance,
        e.g. {{dablink|For the IEEE magazine see [[Computer (magazine)]].}}
        
        """
        
        m = re.search(self.re["disambiguation"], markup)
        if m:
            return self.parse_links(m.group(1))
        else:
            return []
    
    def parse_important(self, markup):
        
        """ Returns a list of words that appear in bold in the article.
        
        Things like table titles are not added to the list,
        these are probably bold because it makes the layout nice,
        not necessarily because they are important.
        
        """
        
        important = []
        table_titles = [table.title for table in self.tables]
        m = re.findall(self.re["bold"], markup)
        for bold in m:
            bold = self.plain(bold)
            if not bold in table_titles:
                important.append(bold.lower())
        
        return important

### DRAWING UTILITIES ################################################################################
   
def is_preformatted(str):

    """ Determines if an item in a paragraph is preformatted.

    If all of the lines in the markup start with a " "
    this indicates preformatted text.
    Preformatted is usually used for programming code.

    """

    for chunk in str.split("\n"):
        if  not chunk.startswith(" "):
            return False
    
    return True

def is_list(str):
 
    """ Determines if an item in a paragraph is a list.

    If all of the lines in the markup start with a "*" or "1." 
    this indicates a list as parsed by parse_paragraphs().
    It can be drawn with draw_list().
    
    """ 
    
    for chunk in str.split("\n"):
        chunk = chunk.replace("\t", "")
        if  not chunk.lstrip().startswith("*") \
        and not re.search(r"^([0-9]{1,3}\. )", chunk.lstrip()):
            return False
    
    return True
    
def is_math(str):
    
    """ Determines if an item in a paragraph is a LaTeX math equation.
    
    Math equations are wrapped in <math></math> tags.
    They can be drawn as an image using draw_math().
    
    """
    
    str = str.strip()
    if str.startswith("<math>") and str.endswith("</math>"):
        return True
    else:
        return False
        
def draw_math(str, x, y, alpha=1.0):
    
    """ Uses mimetex to generate a GIF-image from the LaTeX equation.
    """
    
    try: from web import _ctx
    except: pass
    
    str = re.sub("</{0,1}math>", "", str.strip())
    img = mimetex.gif(str)
    w, h = _ctx.imagesize(img)
    _ctx.image(img, x, y, alpha=alpha)
    return w, h

def textwidth(str):
    
    """textwidth() reports incorrectly when lineheight() is smaller than 1.0
    """
    
    try: from web import _ctx
    except: pass
    
    l = _ctx.lineheight()
    _ctx.lineheight(1)
    w = _ctx.textwidth(str)
    _ctx.lineheight(l)
    
    return w            

def draw_list(markup, x, y, w, padding=5, callback=None):
    
    """ Draws list markup with indentation in NodeBox.

    Draw list markup at x, y coordinates
    using indented bullets or numbers.
    The callback is a command that takes a str and an int.
    
    """
    
    try: from web import _ctx
    except: pass

    i = 1
    for chunk in markup.split("\n"):
        
        if callback != None: 
            callback(chunk, i)
        
        m = re.search("^([0-9]{1,3}\. )", chunk.lstrip())
        if m:
            indent = re.search("[0-9]", chunk).start()*padding*2
            bullet = m.group(1)
            dx = textwidth("000.")
            chunk = chunk.lstrip(m.group(1)+"\t")
        
        if chunk.lstrip().startswith("*"):
            indent = chunk.find("*")*padding*2
            bullet = u"•"
            dx = textwidth("*")
            chunk = chunk.lstrip("* \t")
        
        _ctx.text(bullet, x+indent, y)
        dx += padding + indent
        _ctx.text(chunk, x+dx, y, width=w-dx)
        y += _ctx.textheight(chunk, width=w-dx)
        y += _ctx.textheight(" ") * 0.25
        i += 1

def draw_table(table, x, y, w, padding=5):
    
    """ This is a very poor algorithm to draw Wikipedia tables in NodeBox.
    """
    
    try: from web import _ctx
    except: pass
    
    f = _ctx.fill()
    _ctx.stroke(f)
    h = _ctx.textheight(" ") + padding*2
    
    row_y = y
    
    if table.title != "":
        _ctx.fill(f)
        _ctx.rect(x, row_y, w, h)
        _ctx.fill(1)
        _ctx.text(table.title, x+padding, row_y+_ctx.fontsize()+ padding)
        row_y += h
    
    # A table of flags marking how long a cell 
    # from a previous row is still spanning in a column.
    rowspans = [1 for i in range(10)]
    previous_cell_w = 0
    
    for row in table:
        
        cell_x = x
        
        # The width of a cell is the total table width 
        # evenly divided by the number of cells.
        # Previous rows' cells still spanning will push cells
        # to the right and decrease their width.
        cell_w  = 1.0 * w
        cell_w -= previous_cell_w * len([n for n in rowspans if n > 1])
        cell_w /= len(row)
        
        # The height of each cell is the highest cell in the row.
        # The height depends on the amount of text in the cell.
        cell_h = 0
        for cell in row:
            this_h = _ctx.textheight(cell, width=cell_w-padding*2) + padding*2
            cell_h = max(cell_h, this_h)
        
        # Traverse each cell in this row.
        i = 0
        for cell in row:
            
            # If a previous row's cell is still spanning,
            # push this cell to the right.
            if rowspans[i] > 1:
                rowspans[i] -= 1
                cell_x += previous_cell_w
                i += 1
                
            # Get the rowspan attribute for this cell.
            m = re.search("rowspan=\"(.*?)\"", cell.properties)
            if m:
                rowspan = int(m.group(1))
                rowspans[i] = rowspan
            else:
                rowspan = 1

            # Padded cell text.            
            # Horizontal line above each cell.
            # Vertical line before each cell.
            _ctx.fill(f)
            _ctx.text(cell, cell_x+padding, row_y+_ctx.fontsize()+padding, cell_w-padding*2)
            _ctx.line(cell_x, row_y, cell_x+cell_w, row_y)
            if cell_x > x:
                _ctx.nofill()
                _ctx.line(cell_x, row_y, cell_x, row_y+cell_h)
                
            cell_x += cell_w
            i += 1
            
        # Move to next row.
        row_y += cell_h
        previous_cell_w = cell_w
        
    # Table's bounding rectangle.
    _ctx.nofill()
    _ctx.rect(x, y, w, row_y-y)

### WIKIPEDIASEARCH ##################################################################################

class WikipediaSearch(WikipediaPage, URLAccumulator):
    
    def _api_request(self, q, language="en"):

        url  = "http://"+language+".wikipedia.org/w/api.php"
        url += "?action=query&redirects&format=xml&prop=revisions&rvprop=content&titles="
        url += quote(q)
        return url
    
    def __init__(self, q, language="en", light=False, wait=10, asynchronous=False, cached=True,
                 case_sensitive=False, full_strip=True):
        
        """ A download manager for Wikipedia pages.
        
        WikipediaSearch is a combination of
        URLAccumulator that handles asynchronous and cached web downloads and
        WikipediaPage that parses XML retrieved from the Wikipedia API.
        
        Retrieves the latest revision.
        Redirects are handled by the Wikipedia server.
        
        """
        
        self._light = light
        self._full_strip = full_strip
        
        if cached: 
            cache = "wikipedia"
        else:
            cache = None
        
        if not case_sensitive:
            q = str(q.lower())
        q = q.replace(" ", "_")
        url = self._api_request(q, language)
        URLAccumulator.__init__(self, url, wait, asynchronous, cache, type=".xml", throttle=2)

    def load(self, data):
        
        dom = minidom.parseString(self.data)
        page = dom.getElementsByTagName("page")[0]
        title = page.getAttribute("title")
        try:
            rev = dom.getElementsByTagName("rev")[0]
            data = rev.childNodes[0].nodeValue.strip()
        except:
            if not self.error:
                self.error = WikipediaPageMissing()
            data = ""

        WikipediaPage.__init__(self,  title, data, light=self._light, full_strip=self._full_strip)

def search(q, language="en", light=False, wait=10, asynchronous=False, cached=True, 
           case_sensitive=False, full_strip=True):
    return WikipediaSearch(q, language, light, wait, asynchronous, cached, case_sensitive, full_strip)

######################################################################################################
# Some interesting things...
# Redirects are now handled by the Wikipedia server but for some reason I'm keeping this code around.
# The superscript could be used to format references and footnotes.

def is_redirect(page):
    m = re.search(r"#REDIRECT \[\[.*?\]\]", page)
    if m and len(m.group(0)) == len(page):
        return True
    else:
        return False

def redirect(page):
    m = re.search(r"#REDIRECT \[\[(.*?)\]\]", page)
    if m:
        return m.group(1)
    else:
        return None

def superscript(number):
    
    digits = [
        u"\u2070",
        u"\u2071",
        u"\u2072",
        u"\u2073",
        u"\u2074",
        u"\u2075",
        u"\u2076",
        u"\u2077",
        u"\u2078",
        u"\u2079",  
    ]
    
    s = u""
    for digit in str(number):
        s += digits[int(digit)]
        
    return s

######################################################################################################




### URL ##############################################################################################
# Code for identifying, parsing and retrieving URL's.
# The asynchronous URLAccumulator is subclassed in all other services.

# Author: Tom De Smedt.
# Copyright (c) 2007 by Tom De Smedt.
# See LICENSE.txt for details.

import os
import socket, urllib, urllib2, urlparse
import thread, time
from warnings import warn

from cache import Cache

### SETTINGS #########################################################################################

USER_AGENT = "NodeBox/1.9.4 +http://nodebox.net"
REFERER = "http://nodebox.net"

### URLERROR #########################################################################################

class URLError(Exception):
    # A fault in the URL, like a missing t in htp://
    def __str__(self): return str(self.__class__)  

class URLTimeout(URLError):
    # URL took to long to load.
    def __str__(self): return str(self.__class__)  

class HTTPError(URLError):
    # Error on server.
    def __str__(self): return str(self.__class__)  

class HTTP401Authentication(HTTPError):
    # URL requires a login and password.
    def __str__(self): return str(self.__class__)  

class HTTP403Forbidden(HTTPError):
    # No access to this URL (user-agent?)
    def __str__(self): return str(self.__class__)  
    
class HTTP404NotFound(HTTPError):
    # URL doesn't exist on the internet.
    def __str__(self): return str(self.__class__)  

### URLPARSER ########################################################################################

class URLParser:
    
    def __init__(self, url="", method="get"):
        
        """ Splits an url string into different parts.
        
        The parts are:
        protocol, domain, login, username, password, port, path, page, query, anchor.
        
        The method defaults to get when the url has a query part.
        Setting it to post will submit the query by POST
        when opening the url.
        
        """
        
        # If the url is a URLParser, copy POST parameters correctly.        
        is_post_urlparser = False
        if isinstance(url, URLParser) and url.method == "post":
            is_post_urlparser = True
            url.method = "get"
            
        # If the url is a URLParser, use its string representation.
        # See that the original object's method is correctly reset.
        urlstr = str(url)
        if is_post_urlparser: url.method = "post"
        url = urlstr
        
        # Consider the following url:
        # http://user:pass@example.com:992/animal/bird?species=seagull#wings
        # protocol: http
        # domain: example.com
        url = urlparse.urlsplit(url)
        self.protocol = url[0]
        self.domain = url[1]
        
        # username: user
        # password: pass
        self.username = ""
        self.password = ""
        if self.domain.find("@") >= 0:
            login = self.domain.split("@")[0]
            if login.find(":") >= 0:
                self.username = login.split(":")[0]
                self.password = login.split(":")[1]
            self.domain = self.domain.split("@")[1]
        
        # port: 992
        self.port = ""
        if self.domain.find(":") >= 0:
            p = self.domain.split(":")
            if p[1].isdigit(): 
                self.port = p[1]
                self.domain = p[0]
        
        # path: /animal/
        # page: bird
        self.path = url[2]
        self.page = ""
        if not self.path.endswith("/"):
            if self.path.find("/") >= 0:
                self.page = self.path.split("/")[-1]
                self.path = self.path[:-len(self.page)]
            else:
                self.page = self.path
                self.path = ""
        self.filename = self.page

        # query: {"species": "seagull"}
        self.query = {}
        self.method = method
        if url[3] != "": 
            self.method = "get"
        if is_post_urlparser:
            self.method = "post"
        for param in url[3].split("&"):
            key, value = "", ""
            if param.find("=") >= 0:
                try: (key, value) = param.split("=")
                except: 
                    key = param
            else:
                key = param
            if key != "":
                self.query[key] = value
        
        # anchor: wings
        self.anchor = url[4]
        
    def __str__(self):
        
        """ Reforms a url string from the different parts.
        """
        
        url = ""
        if self.protocol != ""  : url += self.protocol + "://"
        if self.username != ""  : url += self.username + ":" + self.password + "@"
        if self.domain   != ""  : url += self.domain
        if self.port     != ""  : url += ":" + self.port
        if self.path     != ""  : url += self.path
        if self.page     != ""  : url += self.page
        if self.method == "get" and \
           len(self.query) > 0  : url += "?" + urllib.urlencode(self.query)
        if self.anchor   != ""  : url += "#" + self.anchor

        return url
        
    def _address(self):
        return str(self)
    address = property(_address)
        
def parse(url):
    return URLParser(url)
    
def create(url="", method="get"):
    return URLParser(url, method)

#url = parse("http://user:pass@example.com:992/animal/bird?species=seagull#wings")
#print url.domain
#print url.path
#print url.page
#print url

### URL OPENER #######################################################################################

PROXY = None
def set_proxy(host, type="https"):
    global PROXY
    if host != None:
        PROXY = (host, type)
    else:
        PROXY = None

def open(url, wait=10):
    
    """ Returns a connection to a url which you can read().

    When the wait amount is exceeded, raises a URLTimeout.
    When an error occurs, raises a URLError.
    404 errors specifically return a HTTP404NotFound.

    """
    
    # If the url is a URLParser, get any POST parameters.
    post = None
    if isinstance(url, URLParser) and url.method == "post":
        post = urllib.urlencode(url.query)
    
    # If the url is a URLParser (or a YahooResult or something), 
    # use its string representation.
    url = str(url)
    
    # Use urllib instead of urllib2 for local files.
    if os.path.exists(url):
        return urllib.urlopen(url)
 
    else:
        socket.setdefaulttimeout(wait)
        try:
            #connection = urllib2.urlopen(url, post)
            request = urllib2.Request(url, post, {"User-Agent": USER_AGENT, "Referer": REFERER})
            if PROXY:
                p = urllib2.ProxyHandler({PROXY[1]: PROXY[0]})
                o = urllib2.build_opener(p, urllib2.HTTPHandler)
                urllib2.install_opener(o)
            connection = urllib2.urlopen(request)
        except urllib2.HTTPError, e:
            if e.code == 401: raise HTTP401Authentication
            if e.code == 403: raise HTTP403Forbidden
            if e.code == 404: raise HTTP404NotFound
            raise HTTPError
        except urllib2.URLError, e:
            if e.reason[0] == 36: raise URLTimeout
            raise URLError

    return connection

#print open("http://nodebox.net")
#print open("http:/nodebox.net")
#print open("http://ndoebox.net")
#print open("http://nodebox.net/doink")
#print open("url.py").info()
#print open("boink.py").info()
#print open("file://url.py").info()

### URL VALIDATION ###################################################################################

def is_url(url, wait=10):
    
    """ Returns False when no connection can be opened to the url.
    """
    
    try: connection = open(url, wait)
    except:
        return False
        
    return True

def not_found(url, wait=10):
    
    """ Returns True when the url generates a "404 Not Found" error.
    """
    
    try: connection = open(url, wait)
    except HTTP404NotFound:
        return True
    except:
        return False
        
    return False
    
#url = "http://ndoebox.net"
#print is_url(url)
    
#print not_found("http://nodebox.net/nonexistent.html")
#print not_found("http://nodebox.net/")

### URL MIMETYPE ###################################################################################

def is_type(url, types=[], wait=10):

    """ Determine the MIME-type of the document behind the url.
    
    MIME is more reliable than simply checking the document extension.
    Returns True when the MIME-type starts with anything in the list of types.
    
    """

    # Types can also be a single string for convenience.
    if isinstance(types, str):
        types = [types]
    
    try: connection = open(url, wait)
    except:
        return False
        
    type = connection.info()["Content-Type"]
    for t in types:
        if type.startswith(t): return True
        
    return False

def is_webpage(url, wait=10): 
    return is_type(url, "text/html", wait)
is_page = is_webpage
def is_stylesheet(url, wait=10): 
    return is_type(url, "text/css", wait)
def is_plaintext(url, wait=10): 
    return is_type(url, "text/plain", wait)
def is_pdf(url, wait=10): 
    return is_type(url, "application/pdf", wait)
def is_newsfeed(url, wait=10): 
    return is_type(url, ["application/rss+xml", "application/atom+xml"], wait)
def is_image(url, wait=10): 
    return is_type(url, ["image/gif", "image/jpeg", "image/x-png"], wait)
def is_audio(url, wait=10): 
    return is_type(url, ["audio/mpeg", "audio/x-aiff", "audio/x-wav"], wait)
def is_video(url, wait=10): 
    return is_type(url, ["video/mpeg", "video/quicktime"], wait)
def is_archive(url, wait=10): 
    return is_type(url, ["application/x-stuffit", "application/x-tar", "application/zip"], wait)

#print is_webpage("http://nodebox.net")
#print is_archive("http://nodebox.net/code/data/media/coreimage.zip")

### URLACCUMULATOR ###################################################################################

urlaccumulator_throttle = {}

class URLAccumulator:
    
    def __init__(self, url, wait=60, asynchronous=False, cache=None, type=".html", throttle=0):
        
        """ Creates a threaded connection to a url and reads data.
        
        URLAccumulator can run asynchronously which is useful for animations.
        The done property is set to True when downloading is complete.
        The error attribute contains a URLError exception when no data is found.
        
        URLAccumulator data can be cached.
        Downloads that resulted in an error will write an empty file to the cache,
        the data property will be an empty string but no error is logged
        when the data is read from the cache in later calls.
        
        URLAccumulator can be throttled.
        This ensures only a certain amount of requests to a domain
        will happen in a given period of time.
        
        URLAccumulator data is loaded.
        It has a load() method that is called once when done.
        
        """

        self.url = url
        self.data = None
        self.redirect = None
        self.error = None

        if cache != None:
            self.cached = True
            self._cache = Cache(cache, type)
        else:
            self.cached = False
            self._cache = None
        
        self._domain = URLParser(self.url).domain
        self._throttle = throttle
        global urlaccumulator_throttle
        if not self._domain in urlaccumulator_throttle:
            urlaccumulator_throttle[self._domain] = time.time() - self._throttle

        self._start = time.time()
        self._wait = wait
        self._busy = True
        self._loaded = False
 
        # Synchronous downloads wait until completed,
        # otherwise check the done property.
        thread.start_new_thread(self._retrieve, (self.url,))
        if not asynchronous:
            while not self._done():
                time.sleep(0.1)
    
    def _queued(self):

        # Throttles live requests: 
        # waits until the current time is greater than
        # the time of the last request plus the throttle amount.
        global urlaccumulator_throttle   
        if self.cached and self._cache.exists(str(self.url)):
            return False
        elif time.time() < urlaccumulator_throttle[self._domain] + self._throttle:
            return True
        else:
            urlaccumulator_throttle[self._domain] = time.time()
            return False
    
    def _retrieve(self, url):

        # When the url data is stored in cache, load that.
        # Otherwise, retrieve it from the web.
        if self.cached and self._cache.exists(str(url)):
            self.data = self._cache.read(str(url))
        else:
            try: 
                connection = open(url)
                self.data = connection.read()
                self.redirect = connection.geturl()
                if self.redirect == str(url):
                    self.redirect = None
            except Exception, e:
                self.data = u""
                self.error = e
        
        self._busy = False
            
    def _done(self):
        
        # Will continue downloading asynchronously.
        # 1) When the time limit is exceeded, logs a Timeout error.
        # 2) Once uncached data is ready, stores it in cache.
        # 3) Loads the data.
        # 4) Issues a warning when an error occured.
        if (not self._busy or time.time() >= self._start + self._wait) \
        and not self._queued():                                           # 1
            if self.data == None and \
               self.error == None:
                self.data = u""
                self.error = URLTimeout()
                self.load(self.data)
                self._busy = False
            if self.cached and not self._cache.exists(str(self.url)) \
               and self.data != None and self.data != "":                 # 2
                self._cache.write(str(self.url), str(self.data))
            if not self._loaded and self.error == None:                   # 3
                self.load(self.data)
                self._loaded = True
            if self.error != None:                                        # 4
                warn(str(self.error)+" in "+str(self.__class__)+" for "+str(self.url), Warning)
            return True
        else:
            return False
            
    done = property(_done)
    
    def load(self, data):
        
        """ Override this method in subclasses to process downloaded data.
        """
        
        pass

def retrieve(url, wait=60, asynchronous=False, cache=None, type=".html"):
    
    ua = URLAccumulator(url, wait, asynchronous, cache, type)
    return ua

#r = retrieve("http://nodebox.net")
#print r.data
#print r.redirect
            
#url = create("http://api.search.yahoo.com/ContentAnalysisService/V1/termExtraction", method="post")
#url.query["appid"] = "YahooDemo"
#url.query["context"] = "Italian sculptors and painters of the renaissance favored the Virgin Mary for inspiration"
#url.query["query"] = "madonna"
#r = retrieve(url)
#print r.data

#r = retrieve("http://nodebox.net", asynchronous=True)
#while not r.done:
#    print "wait...",
#    time.sleep(0.1)
#print r.redirect



# XXX - should or should we not do quote_plus() somewhere in here?


# BOIDS - last updated for NodeBox 1.9.2
# Author: Tom De Smedt <tomdesmedt@organisms.be>
# Copyright (c) 2007 by Tom De Smedt.
# See LICENSE.txt for details.

# For the original pseucode the algorithm is based on:
# http://www.vergenet.net/~conrad/boids/pseudocode.html



class Boid:
    
    def __init__(self, boids, x, y, z):
        
        self.boids = boids
        self.flock = boids
    
        self.x = 1.0 * x
        self.y = 1.0 * y
        self.z = 1.0 * z
        
        self.vx = 0
        self.vy = 0
        self.vz = 0
        
        self.is_perching = False
        self._perch_t = 0
	        
        
    def copy(self):
        
        b = Boid(self.boids, self.x, self.y, self.z)
        b.vx = self.vx
        b.vy = self.vy
        b.vz = self.vz
        b.is_perching = self.is_perching
        b._perch_t = self._perch_t
        
        return b
        
    def cohesion(self, d=100):
        
        """ Boids move towards the flock's centre of mass.
        
        The centre of mass is the average position of all boids,
        not including itself (the "perceived centre").
        
        """
        
        vx = vy = vz = 0
        for b in self.boids:
            if b != self:
                vx, vy, vz = vx+b.x, vy+b.y, vz+b.z
                
        n = len(self.boids)-1
        vx, vy, vz = vx/n, vy/n, vz/n
        
        return (vx-self.x)/d, (vy-self.y)/d, (vz-self.z)/d
                
    def separation(self, r=10):
        
        """ Boids keep a small distance from other boids.
        
        Ensures that boids don't collide into each other,
        in a smoothly accelerated motion.
        
        """
        
        vx = vy = vz = 0
        for b in self.boids:
            if b != self:
                if abs(self.x-b.x) < r: vx += (self.x-b.x)
                if abs(self.y-b.y) < r: vy += (self.y-b.y)
                if abs(self.z-b.z) < r: vz += (self.z-b.z)
                
        return vx, vy, vz
        
    def alignment(self, d=5):
        
        """ Boids match velocity with other boids.
        """
        
        vx = vy = vz = 0
        for b in self.boids:
           if b != self:
               vx, vy, vz = vx+b.vx, vy+b.vy, vz+b.vz
        
        n = len(self.boids)-1
        vx, vy, vz = vx/n, vy/n, vz/n
        
        return (vx-self.vx)/d, (vy-self.vy)/d, (vz-self.vz)/d
        
    def limit(self, max=30):
        
        """ The speed limit for a boid.
        
        Boids can momentarily go very fast,
        something that is impossible for real animals.
        
        """
        
        if abs(self.vx) > max: 
            self.vx = self.vx/abs(self.vx)*max
        if abs(self.vy) > max: 
            self.vy = self.vy/abs(self.vy)*max
        if abs(self.vz) > max: 
            self.vz = self.vz/abs(self.vz)*max
        
    def _angle(self):
        
        """ Returns the angle towards which the boid is steering.
        """
        
        from math import atan, pi, degrees
        a = degrees(atan(self.vy/self.vx)) + 360
        if self.vx < 0: a += 180

        return a
    
    angle = property(_angle)
        
    def goal(self, x, y, z, d=50.0):
        
        """ Tendency towards a particular place.
        """
        
        return (x-self.x)/d, (y-self.y)/d, (z-self.z)/d
        
class Boids(list):
    
    def __init__(self, n, x, y, w, h):
        
        for i in range(n):
            dx = _ctx.random(w)
            dy = _ctx.random(h)
            z = _ctx.random(200)
            b = Boid(self, x+dx, y+dy, z)
            self.append(b)
            
        self.x = x
        self.y = y
        self.w = w
        self.h = h
        
        self.scattered = False
        self._scatter = 0.005
        self._scatter_t = 50
        self._scatter_i = 0
        
        self._perch = 1.0 # Lower this number to simulate diving.
        self._perch_y = _ctx.HEIGHT
        self._perch_t = lambda:25+_ctx.random(50)
    
        self.has_goal = False
        self.flee = False
        self._gx = 0
        self._gy = 0
        self._gz = 0
	
        
    def copy(self):
        
        boids = Boids(0, self.x, self.y, self.w, self.h)
        
        boids.scattered = self.scattered
        boids._scatter = self._scatter
        boids._scatter_t = self._scatter_t
        boids._scatter_i = self._scatter_i
        
        boids._perch = self._perch
        boids._perch_y = self._perch_y
        boids._perch_t = self._perch_t
        
        boids.has_goal = self.has_goal
        boids.flee = self.flee
        boids._gx = self._gx
        boids._gy = self._gy
        boids._gz = self._gz
        
        for boid in self:
            boids.append(boid.copy())
            
        return boids

    def scatter(self, chance=0.005, frames=50):
        
        self._scatter = chance
        self._scatter_t = frames
    
    def noscatter(self):
        
        self._scatter = 0.0
    
    def perch(self, ground=None, chance=1.0, frames=lambda:25+_ctx.random(50)):
        
        if ground == None:
            ground = _ctx.HEIGHT
            
        self._perch = chance
        self._perch_y = ground
        self._perch_t = frames
    
    def noperch(self):
        
        self._perch = 0.0
    
    def goal(self, x, y, z, flee=False):
        
        self.has_goal = True
        self.flee = flee
        self._gx = x
        self._gy = y
        self._gz = z
        
    def nogoal(self):
        
        self.has_goal = False
        
    def constrain(self):
        
        """ Cages the flock inside the x, y, w, h area.
        
        The actual cage is a bit larger,
        so boids don't seem to bounce of invisible walls
        (they are rather "encouraged" to stay in the area).
        
        If a boid touches the ground level,
        it may decide to perch there for a while.
        
        """
        
        dx = self.w * 0.1
        dy = self.h * 0.1 
        
        for b in self:
            
            if b.x < self.x-dx: b.vx += _ctx.random(dx)
            if b.y < self.y-dy: b.vy += _ctx.random(dy)
            if b.x > self.x+self.w+dx: b.vx -= _ctx.random(dx)
            if b.y > self.y+self.h+dy: b.vy -= _ctx.random(dy)
            if b.z < 0: b.vz += 10
            if b.z > 100: b.vz -= 10
            
            if b.y > self._perch_y and _ctx.random() < self._perch:
                b.y = self._perch_y
                b.vy = -abs(b.vy) * 0.2
                b.is_perching = True
                try:
                    b._perch_t = self._perch_t()
                except:
                    b._perch_t = self._perch_t
            
    def update(self, 
               shuffled=True, 
               cohesion=100, 
               separation=10, 
               alignment=5, 
               goal=20,
               limit=30):
        
        """ Calculates the next motion frame for the flock.
        """
        
        # Shuffling the list of boids ensures fluid movement.
        # If you need the boids to retain their position in the list
        # each update, set the shuffled parameter to False.
        from random import shuffle
        if shuffled: shuffle(self)
        
        m1 = 1.0 # cohesion
        m2 = 1.0 # separation
        m3 = 1.0 # alignment
        m4 = 1.0 # goal
        
        # The flock scatters randomly with a Boids.scatter chance.
        # This means their cohesion (m1) is reversed,
        # and their joint alignment (m3) is dimished,
        # causing boids to oscillate in confusion.
        # Setting Boids.scatter(chance=0) ensures they never scatter.
        if not self.scattered and _ctx.random() < self._scatter:
            self.scattered = True
        if self.scattered:
            m1 = -m1
            m3 *= 0.25
            self._scatter_i += 1
        if self._scatter_i >= self._scatter_t:
            self.scattered = False
            self._scatter_i = 0

        # A flock can have a goal defined with Boids.goal(x,y,z),
        # a place of interest to flock around.
        if not self.has_goal:
            m4 = 0
        if self.flee:
            m4 = -m4
        
        for b in self:
            
            # A boid that is perching will continue to do so
            # until Boid._perch_t reaches zero.
            if b.is_perching:
                if b._perch_t > 0:
                    b._perch_t -= 1
                    continue
                else:
                    b.is_perching = False
            
            vx1, vy1, vz1 = b.cohesion(cohesion)
            vx2, vy2, vz2 = b.separation(separation)
            vx3, vy3, vz3 = b.alignment(alignment)
            vx4, vy4, vz4 = b.goal(self._gx, self._gy, self._gz, goal)
            
            b.vx += m1*vx1 + m2*vx2 + m3*vx3 + m4*vx4
            b.vy += m1*vy1 + m2*vy2 + m3*vy3 + m4*vy4
            b.vz += m1*vz1 + m2*vz2 + m3*vz3 + m4*vz4
            
            b.limit(limit)
        
            b.x += b.vx
            b.y += b.vy
            b.z += b.vz
        
        self.constrain()
        
def flock(n, x, y, w, h):
    
    return Boids(n, x, y, w, h)



#!/usr/bin/env python2

# This file is part of Shoebot.
# Copyright (C) 2007-2009 the Shoebot authors
# See the COPYING file for the full license text.
#
#   Redistribution and use in source and binary forms, with or without
#   modification, are permitted provided that the following conditions are met:
#
#   Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
#   The name of the author may not be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
#   WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
#   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
#   EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
#   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
#   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
#   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
#   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
#   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import os
import cairocffi as cairo

from drawqueue_sink import DrawQueueSink

class CairoImageSink(DrawQueueSink):
    '''
    DrawQueueSink that uses cairo contexts as the render context.
    '''
    def __init__(self, filename=None, format=None, multifile=False, buff=None):
        """
        :param filename:  output filename
        :param format:    if filename is specified this is not needed.
        :param multifile: If used with filename, then numbered files will be output for each froam.
        :param buff:      optionally a file like object can be used instead of a filename
                          this is useful for streaming output.

        """
        DrawQueueSink.__init__(self)
        if format is None:
            if filename is not None:
                format = os.path.splitext(filename)[1][1:].lower()
            elif buff is not None:
                raise AttributeError("No format specified, but using buff")
        self.buff = buff
        self.format = format
        self.filename = filename
        self.multifile = multifile
        self.file_root, self.file_ext = os.path.splitext(filename)

    def _output_file(self, frame):
        """
        If filename was used output a filename, along with multifile
        numbered filenames will be used.

        If buff was specified it is returned.

        :return: Output buff or filename.
        """
        if self.buff:
            return self.buff
        elif self.multifile:
            return self.file_root + "_%03d" % frame + self.file_ext
        else:
            return self.filename

    def create_rcontext(self, size, frame):
        """
        Called when CairoCanvas needs a cairo context to draw on
        """
        if self.format == 'pdf':
            surface = cairo.PDFSurface(self._output_vfile(frame), *size)
        elif self.format == 'png':
            surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, *size)
        elif self.format in ('ps', 'eps'):
            surface = cairo.PSSurface(self._output_file(frame), *size)
        if self.format == 'svg':
            surface = cairo.SVGSurface(self._output_file(frame), *size)
        return cairo.Context(surface)

    def rendering_finished(self, size, frame, cairo_ctx):
        """
        Called when CairoCanvas has rendered a bot
        """
        surface = cairo_ctx.get_target()
        if self.format == 'png':
            surface.write_to_png(self._output_file(frame))
        surface.finish()
        surface.flush()

    def set_title(self, title):
        # Does nothing, only relevant to GUI
        pass



### PAGE #############################################################################################
# Code for querying the HTML DOM.
# It wraps BeautifulSoup by Leonard Richardson.

# Author: Tom De Smedt.
# Copyright (c) 2007 by Tom De Smedt.
# See LICENSE.txt for details.

from BeautifulSoup import BeautifulSoup, Tag

from url import URLAccumulator, URLParser
from html import replace_entities, plain
from cache import Cache

def clear_cache():
    Cache("html").clear()

### PAGE ERRORS ######################################################################################

class PageUnicodeError(Exception):
    def __str__(self): return str(self.__class__)  

class PageParseError(Exception):
    def __str__(self): return str(self.__class__)  

### PAGE #########@###################################################################################

Tag.find_all = Tag.findAll

class Page(BeautifulSoup, URLAccumulator):
    
    """ DOM tree of a HTML page.
    
    Page is essentially an asynchronous download of a BeautifulSoup page.
    It has the following methods:
    description() - returns meta description
    keywords() - returns meta keywords
    links() - by default, returns external links
    find(tag, attribute=value) - find the first tag with given attributes
    find_all(tag, attribute=value) - find all tags with given attributes
    
    find() and find_all() return objects that have find() and find_all() too.
    They're essentially lists of Tag objects.
    
    Alternatively, get tags directly as properties, e.g.
    page.body.p - returns a list of all p Tag objects (each has find() and find_all() )
    
    To get attributes from a Tag:
    p["id"]
    
    """
    
    def __init__(self, url, wait=10, asynchronous=False, cached=True):
        
        if cached: 
            cache = "html"
        else:
            cache = None
        URLAccumulator.__init__(self, url, wait, asynchronous, cache)

    def load(self, data):
        
        data = replace_entities(data)
        try:
            BeautifulSoup.__init__(self, data)
        except UnicodeEncodeError:
            self.error = PageUnicodeError()
            BeautifulSoup.__init__(self, "")
        except:
            self.error = PageParseError()
            BeautifulSoup.__init__(self, "")            

    def _title(self):
        
        """ Returns the page title.
        """    
        
        return self.find("title").string
        
    title = property(_title)    

    def _description(self):
        
        """ Returns the meta description in the page.
        """        

        meta = self.find("meta", {"name":"description"})
        if isinstance(meta, dict) and \
           meta.has_key("content"):
            return meta["content"]
        else:
            return u""
            
    description = property(_description)

    def _keywords(self):
        
        """ Returns the meta keywords in the page.
        """
        
        meta = self.find("meta", {"name":"keywords"})
        if isinstance(meta, dict) and \
           meta.has_key("content"):
            keywords = [k.strip() for k in meta["content"].split(",")]
        else:
            keywords = []
            
        return keywords
        
    keywords = property(_keywords)

    def links(self, external=True):
        
        """ Retrieves links in the page.
        
        Returns a list of URL's.
        By default, only external URL's are returned.
        External URL's starts with http:// and point to another
        domain than the domain the page is on.
        
        """
        
        domain = URLParser(self.url).domain
        
        links = []
        for a in self("a"):
            for attribute, value in a.attrs:
                if attribute == "href":
                    if not external \
                    or (value.startswith("http://") and value.find("http://"+domain) < 0):
                        links.append(value)
                        
        return links
    
    def find_class(self, classname, tag=""):
        return self( tag, {"class": classname} )
        
def parse(url, wait=10, asynchronous=False, cached=True):
    return Page(url, wait, asynchronous, cached)

"""
import url
url = url.create("http://nodebox.net/code/index.php/Share")
url.query["p"] = 2
print url

page = parse(url)
print page.title
print page.title.string
print page.description()
print page.keywords()

print page.find(id="content")["id"]
# find() returns a list of Tags and has a find() method
for p in page.body.find("div", id="content").find_all("p"):
    print ">>>", plain(p)

print page.links()
print page.find_all("h2")

print page.contents[0].name

# .div returns a list of Tags
print page.body.div(id="content")[0].p
"""


# Copyright (c) 2007 Tom De Smedt.
# See LICENSE.txt for details.

from types import FunctionType, LambdaType

#--- LIST OPERATIONS ---------------------------------------------------------------------------------

def sorted(list, cmp=None, reversed=False):
    """ Returns a sorted copy of the list.
    """
    list = [x for x in list]
    list.sort(cmp)
    if reversed: list.reverse()
    return list

def unique(list):
    """ Returns a copy of the list without duplicates.
    """
    unique = []; [unique.append(x) for x in list if x not in unique]
    return unique    

#--- SET THEORY --------------------------------------------------------------------------------------

def flatten(node, distance=1):
    
    """ Recursively lists the node and its links.
    
    Distance of 0 will return the given [node].
    Distance of 1 will return a list of the node and all its links.
    Distance of 2 will also include the linked nodes' links, etc.
    
    """
    
    # When you pass a graph it returns all the node id's in it.
    if hasattr(node, "nodes") and hasattr(node, "edges"):
        return [n.id for n in node.nodes]
    
    all = [node]
    if distance >= 1:
        for n in node.links: 
            all += n.flatten(distance-1)
    
    return unique(all)
    
def intersection(a, b):
    """ Returns the intersection of lists.
    a & b -> elements that appear in a as well as in b.
    """
    return filter(lambda x: x in a, b)
    
    
def union(a, b):
    """ Returns the union of lists.
    a | b -> all elements from a and all the elements from b.
    """     
    return a + filter(lambda x: x not in a, b)

def difference(a, b):
    """ Returns the difference of lists.
    a - b -> elements that appear in a but not in b.
    """
    return filter(lambda x: x not in b, a)
    
#--- SUBGRAPH ----------------------------------------------------------------------------------------

def subgraph(graph, id, distance=1):
    
    """ Creates the subgraph of the flattened node with given id (or list of id's).
    Finds all the edges between the nodes that make up the subgraph.
    """
    
    g = graph.copy(empty=True)
    
    if isinstance(id, (FunctionType, LambdaType)):
        # id can also be a lambda or function that returns True or False
        # for each node in the graph. We take the id's of nodes that pass.
        id = [node.id for node in filter(id, graph.nodes)]
    if not isinstance(id, (list, tuple)):
        id = [id]
    for id in id:
        for n in flatten(graph[id], distance):
            g.add_node(n.id, n.r, n.style, n.category, n.label, (n==graph.root), n.__dict__)
        
    for e in graph.edges:
        if g.has_key(e.node1.id) and \
           g.has_key(e.node2.id):
            g.add_edge(e.node1.id, e.node2.id, e.weight, e.length, e.label, e.__dict__)
    
    # Should we look for shortest paths between nodes here?
    
    return g
 
#--- CLIQUE ----------------------------------------------------------------------------------------
 
def is_clique(graph):
    
    """ A clique is a set of nodes in which each node is connected to all other nodes.
    """
    
    #for n1 in graph.nodes:
    #    for n2 in graph.nodes:
    #        if n1 != n2 and graph.edge(n1.id, n2.id) == None:
    #            return False

    if graph.density < 1.0: 
        return False
    
    return True
    
def clique(graph, id):
    
    """ Returns the largest possible clique for the node with given id.
    """
    
    clique = [id]
    for n in graph.nodes:
        friend = True
        for id in clique:
            if n.id == id or graph.edge(n.id, id) == None:
                friend = False
                break
        if friend:
            clique.append(n.id)
    
    return clique
    
def cliques(graph, threshold=3):
    
    """ Returns all the cliques in the graph of at least the given size.
    """
    
    cliques = []
    for n in graph.nodes:
        c = clique(graph, n.id)
        if len(c) >= threshold: 
            c.sort()
            if c not in cliques:
                cliques.append(c)
    
    return cliques

#--- UNCONNECTED SUBGRAPHS -------------------------------------------------------------------------

def partition(graph):
    
    """ Splits unconnected subgraphs.
    
    For each node in the graph, make a list of its id and all directly connected id's.
    If one of the nodes in this list intersects with a subgraph,
    they are all part of that subgraph.
    Otherwise, this list is part of a new subgraph.
    Return a list of subgraphs sorted by size (biggest-first).
    
    """
    
    g = []
    for n in graph.nodes:
        c = [n.id for n in flatten(n)]
        f = False
        for i in range(len(g)):
            if len(intersection(g[i], c)) > 0:
                g[i] = union(g[i], c)
                f = True
                break
        if not f:
            g.append(c)
    
    # If 1 is directly connected to 2 and 3,
    # and 4 is directly connected to 5 and 6, these are separate subgraphs.
    # If we later find that 7 is directly connected to 3 and 6,
    # it will be attached to [1, 2, 3] yielding
    # [1, 2, 3, 6, 7] and [4, 5, 6].
    # These two subgraphs are connected and need to be merged.
    merged = []
    for i in range(len(g)):
        merged.append(g[i])
        for j in range(i+1, len(g)):
            if len(intersection(g[i], g[j])) > 0:
                merged[-1].extend(g[j])
                g[j] = []
    
    g = merged
    g = [graph.sub(g, distance=0) for g in g]
    g.sort(lambda a, b: len(b) - len(a))             
    
    return g


# PhotoBot 0.8 beta - last updated for NodeBox 1rc4 
# Author: Tom De Smedt <tomdesmedt@trapdoor.be>
# Manual: http://nodebox.net/code/index.php/PhotoBot
# Copyright (c) 2006 by Tom De Smedt.
# Refer to the "Use" section on http://nodebox.net/code/index.php/Use

from PIL import Image, ImageChops, ImageFilter, ImageEnhance, ImageOps, ImageDraw, ImageStat

NEAREST = Image.NEAREST
BICUBIC = Image.BICUBIC
BILINEAR = Image.BILINEAR
INTERPOLATION = BILINEAR

LAYERS = []

NORMAL = "normal"
MULTIPLY = "multiply"
SCREEN = "screen"
OVERLAY = "overlay"
HUE = "hue"
COLOR = "color"

HORIZONTAL = "horizontal"
VERTICAL = "vertical"

LINEAR = "linear"
RADIAL = "radial"
DIAMOND = "diamond"

class Canvas:
    
    """Implements a canvas with layers.
    
    A canvas is an empty Photoshop document,
    where layers can be placed and manipulated.
    
    """

    def __init__(self, w, h):
        
        """Creates a new canvas.
    
        Creates the working area on which to blend layers.
        The canvas background is transparent,
        but a background color could be set using the fill() function.
    
        """
        
        self.interpolation = BILINEAR
        self.layers = Layers()
        self.w = w
        self.h = h
        
        img = Image.new("RGBA", (w,h), (255,255,255,0))
        self.layer(img, name="_bg")
        
    def layer(self, img, x=0, y=0, name=""):
    
        """Creates a new layer from file, Layer, PIL Image.
    
        If img is an image file or PIL Image object,
        Creates a new layer with the given image file.
        The image is positioned on the canvas at x, y.
        
        If img is a Layer,
        uses that layer's x and y position and name.
    
        """

        from types import StringType
        if isinstance(img, Image.Image):
            img = img.convert("RGBA")
            self.layers.append(Layer(self, img, x, y, name))
            return len(self.layers)-1
        if isinstance(img, Layer):
            img.canvas = self
            self.layers.append(img)
            return len(self.layers)-1                 
        if type(img) == StringType: 
            img = Image.open(img)
            img = img.convert("RGBA")
            self.layers.append(Layer(self, img, x, y, name))
            return len(self.layers)-1
    
    def fill(self, rgb, x=0, y=0, w=None, h=None, name=""):
    
        """Creates a new fill layer.
    
        Creates a new layer filled with the given rgb color.
        For example, fill((255,0,0)) creates a red fill.
        The layers fills the entire canvas by default.
    
        """ 
    
        if w == None: w = self.w - x
        if h == None: h = self.h - y
        img = Image.new("RGBA", (w,h), rgb)
        self.layer(img, x, y, name)
        
    def gradient(self, style=LINEAR, w=1.0, h=1.0, name=""):
    
        """Creates a gradient layer.
    
        Creates a gradient layer, that is usually used
        together with the mask() function.
    
        All the image functions work on gradients,
        so they can easily be flipped, rotated, scaled, inverted,
        made brighter or darker, ...
    
        Styles for gradients are LINEAR, RADIAL and DIAMOND.
    
        """
    
        from types import FloatType
        w0 = self.w 
        h0 = self.h
        if type(w) == FloatType: w *= w0
        if type(h) == FloatType: h *= h0
    
        img = Image.new("L", (int(w),int(h)), 255)
        draw = ImageDraw.Draw(img)
    
        if style == LINEAR:
            for i in range(int(w)):
                k = 255.0 * i/w
                draw.rectangle((i, 0, i, h), fill=int(k))
            
        if style == RADIAL:
            r = min(w,h)/2
            for i in range(int(r)):
                k = 255 - 255.0 * i/r
                draw.ellipse((w/2-r+i, h/2-r+i, w/2+r-i, h/2+r-i), fill=int(k))
            
        if style == DIAMOND:
            r = max(w,h)
            for i in range(int(r)):
                x = int(i*w/r*0.5)
                y = int(i*h/r*0.5)
                k = 255.0 * i/r
                draw.rectangle((x, y, w-x, h-y), outline=int(k))
    
        img = img.convert("RGBA")
        self.layer(img, 0, 0, name="")
        
    def merge(self, layers):
        
        """Flattens the given layers on the canvas.
        
        Merges the given layers with the indices in the list
        on the bottom layer in the list.
        The other layers are discarded.
        
        """
        
        layers.sort()
        if layers[0] == 0: del layers[0]
        self.flatten(layers)
        
    def flatten(self, layers=[]):
    
        """Flattens all layers according to their blend modes.
    
        Merges all layers to the canvas,
        using the blend mode and opacity defined for each layer.
        Once flattened, the stack of layers is emptied except
        for the transparent background (bottom layer).
    
        """
        
        #When the layers argument is omitted,
        #flattens all the layers on the canvas.
        #When given, merges the indexed layers.
        
        #Layers that fall outside of the canvas are cropped:
        #this should be fixed by merging to a transparent background
        #large enough to hold all the given layers' data
        #(=time consuming).
        
        if layers == []: 
            layers = range(1, len(self.layers))
        background = self.layers._get_bg()
        background.name = "Background"
        
        for i in layers:

            layer = self.layers[i]
        
            #Determine which portion of the canvas
            #needs to be updated with the overlaying layer.
        
            x = max(0, layer.x)
            y = max(0, layer.y)
            w = min(background.w, layer.x+layer.w)
            h = min(background.h, layer.y+layer.h)
        
            base = background.img.crop((x, y, w, h))

            #Determine which piece of the layer
            #falls within the canvas.

            x = max(0, -layer.x)
            y = max(0, -layer.y)
            w -= layer.x
            h -= layer.y

            blend = layer.img.crop((x, y, w, h))
        
            #Buffer layer blend modes:
            #the base below is a flattened version
            #of all the layers below this one,
            #on which to merge this blended layer.
        
            if layer.blend == NORMAL:
                buffer = blend
            if layer.blend == MULTIPLY:
                buffer = ImageChops.multiply(base, blend)
            if layer.blend == SCREEN:
                buffer = ImageChops.screen(base, blend)
            if layer.blend == OVERLAY:
                buffer = Blend().overlay(base, blend)
            if layer.blend == HUE:
                buffer = Blend().hue(base, blend)
            if layer.blend == COLOR:
                buffer = Blend().color(base, blend)
            
            #Buffer a merge between the base and blend
            #according to the blend's alpha channel:
            #the base shines through where the blend is less opaque.
        
            #Merging the first layer to the transparent canvas
            #works slightly different than the other layers.
        
            alpha = buffer.split()[3]
            if i == 1:
                buffer = Image.composite(base, buffer, base.split()[3])
            else:
                buffer = Image.composite(buffer, base, alpha)
        
            #The alpha channel becomes a composite of
            #this layer and the base:
            #the base's (optional) tranparent background
            #is retained in arrays where the blend layer
            #is transparent as well.
        
            alpha = ImageChops.lighter(alpha, base.split()[3])
            buffer.putalpha(alpha)
        
            #Apply the layer's opacity,
            #merging the buffer to the base with
            #the given layer opacity.
        
            base = Image.blend(base, buffer, layer.alpha)

            #Merge the base to the flattened canvas.

            x = max(0, layer.x)
            y = max(0, layer.y)
            background.img.paste(base, (x,y))
    
        layers.reverse()
        for i in layers: del self.layers[i]
    
        img = Image.new("RGBA", (self.w,self.h), (255,255,255,0))
        self.layers._set_bg(Layer(self, img, 0, 0, name="_bg"))
        
        if len(self.layers) == 1:
            self.layers.append(background)
        else:
            self.layers.insert(layers[-1], background)
        
    def export(self, filename):
    
        """Exports the flattened canvas.
    
        Flattens the canvas.
        PNG retains the alpha channel information.
        Other possibilities are JPEG and GIF.
    
        """

        self.flatten()
        self.layers[1].img.save(filename)
        return filename
        
    def draw(self, x, y):
        
        """Places the flattened canvas in NodeBox.
        
        Exports to a temporary PNG file.
        Draws the PNG in NodeBox using the image() command.
        Removes the temporary file.
        
        """
        
        try:
            from time import time
            import md5
            from os import unlink
            m = md5.new()
            m.update(str(time()))
            filename = "photobot" + str(m.hexdigest()) + ".png"
            self.export(filename)
            _ctx.image(filename, x, y)
            unlink(filename)
        
        except:
            pass
    
    def preferences(interpolation=BILINEAR):
    
        """Settings that influence image manipulation.
    
        Currently, only defines the image interpolation,
        which can be set to NEAREST, BICUBIC or BILINEAR.
    
        """
    
        self. interpolation = interpolation
        
def canvas(w, h):
    
    return Canvas(w, h)
    
class Layers(list):
    
    """Extends the canvas.layers[] list so it indexes layers names.
    
    When the index is an integer, returns the layer at that  index.
    When the index is a string, returns the first layer with that name.
    
    The first element, canvas.layers[0],
    is the transparent background and must remain untouched.
    
    """
    
    def __getitem__(self, index):

        from types import StringType
        if type(index) == StringType:
            for layer in self:
                if layer.name == index: return layer
        else:
            return list.__getitem__(self, index)
        
    def _get_bg(self):
        
        return list.__getitem__(self, 0)
        
    def _set_bg(self, layer):
        
        list.__setitem__(self, 0, layer)

class Layer:
    
    """Implements a layer on the canvas.
    
    A canvas layer stores an image at a given position on the canvas,
    and all the Photoshop transformations possible for this layer:
    duplicate(), desature(), overlay(), rotate(), and so on.
    
    """
    
    def __init__(self, canvas, img, x=0, y=0, name=""):
        
        self.canvas = canvas
        self.name = name
        self.img = img
        self.x = x
        self.y = y
        self.w = img.size[0]
        self.h = img.size[1]
        self.alpha = 1.0
        self.blend = NORMAL
        self.pixels = Pixels(self.img, self)
        
    def index(self):
        
        """Returns this layer's index in the canvas.layers[].
        
        Searches the position of this layer in the canvas'
        layers list, return None when not found.
        
        """
        
        for i in range(len(self.canvas.layers)):
            if self.canvas.layers[i] == self: break
        if self.canvas.layers[i] == self: 
            return i
        else:
            return None
            
    def copy(self):
        
        """Returns a copy of the layer.
        
        This is different from the duplicate() method,
        which duplicates the layer as a new layer on the canvas.
        The copy() method returns a copy of the layer
        that can be added to a different canvas.
        
        """
        
        layer = Layer(None, self.img.copy(), self.x, self.y, self.name)
        layer.w = self.w
        layer.h = self.h
        layer.alpha = self.alpha
        layer.blend = self.blend
        
        return layer
        
    def delete(self):
        
        """Removes this layer from the canvas.
              
        """
        
        i = self.index()
        if i != None: del self.canvas.layers[i]
        
    def up(self):
        
        """Moves the layer up in the stacking order.
        
        """
        
        i = self.index()
        if i != None:
            del self.canvas.layers[i]
            i = min(len(self.canvas.layers), i+1)
            self.canvas.layers.insert(i, self)
            
    def down(self):
        
        """Moves the layer down in the stacking order.
        
        """
        
        i = self.index()
        if i != None:
            del self.canvas.layers[i]
            i = max(0, i-1)
            self.canvas.layers.insert(i, self)
    
    def bounds(self):
    
        """Returns the size of the layer.
    
        This is the width and height of the bounding box,
        the invisible rectangle around the layer.
    
        """
    
        return self.img.size
    
    def select(self, path, feather=True):
    
        """Applies the polygonal lasso tool on a layer.
    
        The path paramater is a list of points,
        either [x1, y1, x2, y2, x3, y3, ...]
        or [(x1,y1), (x2,y2), (x3,y3), ...]
    
        The parts of the layer that fall outside
        this polygonal area are cut.
        
        The selection is not anti-aliased,
        but the feather parameter creates soft edges.
    
        """
    
        w, h = self.img.size
        mask = Image.new("L", (w,h), 0)
        draw = ImageDraw.Draw(mask)
        
        draw = ImageDraw.Draw(mask)
        draw.polygon(path, fill=255)

        if feather:
            mask = mask.filter(ImageFilter.SMOOTH_MORE)
            mask = mask.filter(ImageFilter.SMOOTH_MORE)
            
        mask = ImageChops.darker(mask, self.img.split()[3])
        self.img.putalpha(mask)
    
    def mask(self):
    
        """Masks the layer below with this layer.
    
        Commits the current layer to the alpha channel of 
        the previous layer. Primarily, mask() is useful when 
        using gradient layers as masks on images below. 
    
        For example:
        canvas.layer("image.jpg")
        canvas.gradient()
        canvas.layer(2).flip()
        canvas.layer(2).mask()
    
        Adds a white-to-black linear gradient to
        the alpha channel of image.jpg, 
        making it evolve from opaque on 
        the left to transparent on the right.
    
        """

        if len(self.canvas.layers) < 2: return
        i = self.index()
        if i == 0: return
        
        layer = self.canvas.layers[i-1]
    
        alpha = Image.new("L", layer.img.size, 0)
    
        #Make a composite of the mask layer in grayscale
        #and its own alpha channel.
    
        mask = self.canvas.layers[i]        
        flat = ImageChops.darker(mask.img.convert("L"), mask.img.split()[3])
        alpha.paste(flat, (mask.x,mask.y))
        alpha = ImageChops.darker(alpha, layer.img.split()[3])
        layer.img.putalpha(alpha)
    
        self.delete()
    
    def duplicate(self):
    
        """Creates a copy of the current layer.
    
        This copy becomes the top layer on the canvas.
    
        """
    
        i = self.canvas.layer(self.img.copy(), self.x, self.y, self.name)
        clone = self.canvas.layers[i]
        clone.alpha = self.alpha
        clone.blend = self.blend
                    
    def opacity(self, a=100):
    
        self.alpha = a * 0.01
    
    def multiply(self):
    
        self.blend = MULTIPLY
    
    def screen(self):
    
        self.blend = SCREEN
    
    def overlay(self):

        self.blend = OVERLAY
        
    def hue(self):
    
        self.blend = HUE
        
    def color(self):
    
        self.blend = COLOR
        
    def brightness(self, value=1.0):

        """Increases or decreases the brightness in the layer.
    
        The given value is a percentage to increase
        or decrease the image brightness,
        for example 0.8 means brightness at 80%.
    
        """
     
        b = ImageEnhance.Brightness(self.img) 
        self.img = b.enhance(value)
    
    def contrast(self, value=1.0):
    
        """Increases or decreases the contrast in the layer.
    
        The given value is a percentage to increase
        or decrease the image contrast,
        for example 1.2 means contrast at 120%.
    
        """

        c = ImageEnhance.Contrast(self.img) 
        self.img = c.enhance(value) 
    
    def desaturate(self):
    
        """Desaturates the layer, making it grayscale.
    
        Instantly removes all color information from the layer,
        while maintaing its alpha channel.
    
        """
    
        alpha = self.img.split()[3]
        self.img = self.img.convert("L")
        self.img = self.img.convert("RGBA")
        self.img.putalpha(alpha)
    
    def invert(self):
    
        """Inverts the layer.
    
        """
    
        alpha = self.img.split()[3]
        self.img = self.img.convert("RGB")
        self.img = ImageOps.invert(self.img)
        self.img = self.img.convert("RGBA")
        self.img.putalpha(alpha)
    
    def translate(self, x, y):
    
        """Positions the layer at the given coordinates.
    
        The x and y parameters define where to position 
        the top left corner of the layer,
        measured from the top left of the canvas.
    
        """
    
        self.x = x
        self.y = y
    
    def scale(self, w=1.0, h=1.0):
    
        """Resizes the layer to the given width and height.
    
        When width w or height h is a floating-point number,
        scales percentual, 
        otherwise scales to the given size in pixels.
    
        """

        from types import FloatType
        w0, h0 = self.img.size
        if type(w) == FloatType: w = int(w*w0)
        if type(h) == FloatType: h = int(h*h0)
    
        self.img = self.img.resize((w,h), INTERPOLATION)
        self.w = w
        self.h = h
    
    def distort(self, x1=0,y1=0, x2=0,y2=0, x3=0,y3=0, x4=0,y4=0):
    
        """Distorts the layer.
        
        Distorts the layer by translating 
        the four corners of its bounding box to the given coordinates:
        upper left (x1,y1), upper right(x2,y2),
        lower right (x3,y3) and lower left (x4,y4).
        
        """
    
        w, h = self.img.size
        quad = (-x1,-y1, -x4,h-y4, w-x3,w-y3, w-x2,-y2)
        self.img = self.img.transform(self.img.size, Image.QUAD, quad, INTERPOLATION)    

    def rotate(self, angle):
    
        """Rotates the layer.
    
        Rotates the layer by given angle.
        Positive numbers rotate counter-clockwise,
        negative numbers rotate clockwise.
    
        Rotate commands are executed instantly,
        so many subsequent rotates will distort the image.
    
        """
    
        #When a layer rotates, its corners will fall outside
        #of its defined width and height.
        #Thus, its bounding box needs to be expanded.
    
        #Calculate the diagonal width, and angle from the layer center.
        #This way we can use the layers's corners 
        #to calculate the bounding box.
    
        from math import sqrt, pow, sin, cos, degrees, radians, asin
        w0, h0 = self.img.size
        d = sqrt(pow(w0,2) + pow(h0,2))
        d_angle = degrees(asin((w0*0.5) / (d*0.5)))
    
        angle = angle % 360
        if angle > 90 and angle <= 270: d_angle += 180
    
        w = sin(radians(d_angle + angle)) * d
        w = max(w, sin(radians(d_angle - angle)) * d)
        w = int(abs(w))
    
        h = cos(radians(d_angle + angle)) * d
        h = max(h, cos(radians(d_angle - angle)) * d)
        h = int(abs(h))
    
        dx = int((w-w0) / 2)
        dy = int((h-h0) / 2)
        d = int(d)

        #The rotation box's background color
        #is the mean pixel value of the rotating image.
        #This is the best option to avoid borders around
        #the rotated image.

        bg = ImageStat.Stat(self.img).mean
        bg = (int(bg[0]), int(bg[1]), int(bg[2]), 0)

        box = Image.new("RGBA", (d,d), bg)
        box.paste(self.img, ((d-w0)/2, (d-h0)/2))
        box = box.rotate(angle, INTERPOLATION)
        box = box.crop(((d-w)/2+2, (d-h)/2, d-(d-w)/2, d-(d-h)/2))
        self.img = box
    
        #Since rotate changes the bounding box size,
        #update the layers' width, height, and position,
        #so it rotates from the center.
    
        self.x += (self.w-w)/2
        self.y += (self.h-h)/2
        self.w = w
        self.h = h   
    
    def flip(self, axis=HORIZONTAL):
    
        """Flips the layer, either HORIZONTAL or VERTICAL.
    
        """

        if axis == HORIZONTAL:
            self.img = self.img.transpose(Image.FLIP_LEFT_RIGHT)
        if axis == VERTICAL:
            self.img = self.img.transpose(Image.FLIP_TOP_BOTTOM)
        
    def blur(self):
        
        """Blurs the layer.
        
        """
    
        self.img = self.img.filter(ImageFilter.BLUR)

    def sharpen(self, value=1.0):

        """Increases or decreases the sharpness in the layer.
    
        The given value is a percentage to increase
        or decrease the image sharpness,
        for example 0.8 means sharpness at 80%.
    
        """
     
        s = ImageEnhance.Sharpness(self.img) 
        self.img = s.enhance(value)
        
    def statistics(self):
        
        return ImageStat.Stat(self.img, self.img.split()[3])
        
    def levels(self):
        
        """Returns a histogram for each RGBA channel.
        
        Returns a 4-tuple of lists, r, g, b, and a.
        Each list has 255 items, a count for each pixel value.
                
        """
        
        h = self.img.histogram()
        r = h[0:255]
        g = h[256:511]
        b = h[512:767]
        a = h[768:1024]
        
        return r, g, b, a
    
class Blend:
    
    """Layer blending modes.
    
    Implements additional blending modes to those present in PIL.
    These blending functions can not be used separately from
    the canvas.flatten() method, where the alpha compositing
    of two layers is handled.
    
    Since these blending are not part of a C library,
    but pure Python, they take forever to process.
    
    """
    
    def overlay(self, img1, img2):
    
        """Applies the overlay blend mode.
    
        Overlays image img2 on image img1.
        The overlay pixel combines multiply and screen:
        it multiplies dark pixels values and screen light values.
        Returns a composite image with the alpha channel retained.
    
        """
    
        p1 = list(img1.getdata())
        p2 = list(img2.getdata())
    
        for i in range(len(p1)):
        
            p3 = ()
            for j in range(len(p1[i])):

                a = p1[i][j] / 255.0
                b = p2[i][j] / 255.0
            
                #When overlaying the alpha channels,
                #take the alpha of the most transparent layer.
            
                if j == 3:
                    #d = (a+b)*0.5
                    #d = a
                    d = min(a,b)
                elif a > 0.5: 
                    d = 2*(a+b-a*b)-1
                else: 
                    d = 2*a*b            
                p3 += (int(d*255),)
        
            p1[i] = p3
        
        img = Image.new("RGBA", img1.size, 255)
        img.putdata(p1)
        return img
    
    def hue(self, img1, img2):
    
        """Applies the hue blend mode.
    
        Hues image img1 with image img2.
        The hue filter replaces the hues of pixels in img1
        with the hues of pixels in img2.
        Returns a composite image with the alpha channel retained.
    
        """

        import colorsys

        p1 = list(img1.getdata())
        p2 = list(img2.getdata())
        for i in range(len(p1)):
        
            r1, g1, b1, a1 = p1[i]
            r1 = r1 / 255.0
            g1 = g1 / 255.0
            b1 = b1 / 255.0
        
            h1, s1, v1 = colorsys.rgb_to_hsv(r1, g1, b1)
        
            r2, g2, b2, a2 = p2[i]
            r2 = r2 / 255.0
            g2 = g2 / 255.0
            b2 = b2 / 255.0
            h2, s2, v2 = colorsys.rgb_to_hsv(r2, g2, b2)
        
            r3, g3, b3 = colorsys.hsv_to_rgb(h2, s1, v1)
        
            r3 = int(r3*255)
            g3 = int(g3*255)
            b3 = int(b3*255)
            p1[i] = (r3, g3, b3, a1)
    
        img = Image.new("RGBA", img1.size, 255)
        img.putdata(p1)
        return img
    
    def color(self, img1, img2):
    
        """Applies the color blend mode.
    
        Colorize image img1 with image img2.
        The color filter replaces the hue and saturation of pixels in img1
        with the hue and saturation of pixels in img2.
        Returns a composite image with the alpha channel retained.
    
        """

        import colorsys

        p1 = list(img1.getdata())
        p2 = list(img2.getdata())
        for i in range(len(p1)):
        
            r1, g1, b1, a1 = p1[i]
            r1 = r1 / 255.0
            g1 = g1 / 255.0
            b1 = b1 / 255.0
        
            h1, s1, v1 = colorsys.rgb_to_hsv(r1, g1, b1)
        
            r2, g2, b2, a2 = p2[i]
            r2 = r2 / 255.0
            g2 = g2 / 255.0
            b2 = b2 / 255.0
            h2, s2, v2 = colorsys.rgb_to_hsv(r2, g2, b2)
        
            r3, g3, b3 = colorsys.hsv_to_rgb(h2, s2, v1)
        
            r3 = int(r3*255)
            g3 = int(g3*255)
            b3 = int(b3*255)
            p1[i] = (r3, g3, b3, a1)
    
        img = Image.new("RGBA", img1.size, 255)
        img.putdata(p1)
        return img

class Pixels:
    
    """Provides direct access to a layer's pixels.
    
    The layer.pixels[] contains all pixel values
    in a 1-dimensional array.
    Each pixel is a tuple containing (r,g,b,a) values.
    
    After the array has been updated, layer.pixels.update()
    must be called for the changes to commit.
    
    """
    
    def __init__(self, img, layer):
        
        self.layer = layer
        self.img = img
        self.data = None
        
    def __getitem__(self, i):

        w, h = self.img.size
        if i >= w*h: i -= w*h
        if i < 0: i += w*h
        
        if self.data == None: self.data = list(self.img.getdata())
        return self.data[i]
        
    def __setitem__(self, i, rgba):
        
        w, h = self.img.size
        if i >= w*h: i -= w*h
        if i < 0: i += w*h
        
        if self.data == None: self.data = list(self.img.getdata())
        self.data[i] = rgba

    def __iter__(self):
        
        for i in range(len(self)):
            yield self[i]

    def __len__(self):
        
        w, h = self.img.size
        return w * h
                    
    def update(self):
        
        if self.data != None: 
            self.img.putdata(self.data)
            self.data = None
        
    def convolute(self, kernel, scale=None, offset=0):
        
        """A (3,3) or (5,5) convolution kernel.
        
        The kernel argument is a list with either 9 or 25 elements,
        the weight for each surrounding pixels to convolute.
        
        """
        
        if len(kernel)   ==  9: size = (3,3)
        elif len(kernel) == 25: size = (5,5)
        else: return
        
        if scale == None:
            scale = 0
            for x in kernel: scale += x
            if scale == 0: scale = 1
     
        f = ImageFilter.BuiltinFilter()
        f.filterargs = size, scale, offset, kernel
        self.layer.img = self.layer.img.filter(f)



class InputDeviceMixin(object):
    SUPPORTED_CALLBACKS = 'key_pressed', 'key_released', 'mouse_button_down', 'mouse_button_up', 'mouse_pointer_moved'

    def __init__(self, **kwargs):
        def nop(*args):
            pass

        for name in self.SUPPORTED_CALLBACKS:
            func = kwargs.get(name, nop)
            setattr(self, name, func)

        # Implementation needs to manipulate keys_pressed, mouse_buttons_down
        self.keys_pressed = set()
        self.mouse_buttons_down = set()


    def set_callbacks(self, **kwargs):
        ''' Set callbacks for input events '''
        for name in self.SUPPORTED_CALLBACKS:
            func = kwargs.get(name, getattr(self, name))
            setattr(self, name, func)

    def get_key_down(self):
        ''' Return True if any key is pressed '''
        return bool(self.keys_pressed)

    def get_mouse_down(self):
        ''' Return True if any mouse button is pressed '''
        return bool(self.mouse_buttons_down)
        
    def get_key_map(self):
        return {}

    key_down = property(get_key_down)
    mouse_down = property(get_mouse_down)

            




# Shoebot events
#
# Using python pubsub
#
# Currently VERY simple, just passing when to QUIT and RELOAD_SOURCE
#
# In future a lot of the could probably be simplified by using more events
# and less flags etc.

import collections
import pubsub

channel = pubsub.subscribe("shoebot")

def namedtuple_with_defaults(typename, field_names, default_values=()):
    T = collections.namedtuple(typename, field_names)
    T.__new__.__defaults__ = (None,) * len(T._fields)
    if isinstance(default_values, collections.Mapping):
        prototype = T(**default_values)
    else:
        prototype = T(*default_values)
    T.__new__.__defaults__ = tuple(prototype)
    return T

Event = namedtuple = namedtuple_with_defaults("Event", "type data", dict(data=None))

QUIT_EVENT = "quit"
SOURCE_CHANGED_EVENT = "source-changed"
EVENT_VARIABLE_UPDATED = "variable-updated"
SET_WINDOW_TITLE = "set-window-title"

def next_event(block=False, timeout=None):
    """
    :param block:
    :param timeout:
    :return: event, data
    """
    try:
        return channel.listen(block=block, timeout=timeout).next()['data']
    except StopIteration:
        return None

def event_is(event, event_t):
    return event is not None and event.type == event_t

def publish_event(event_t, data=None, extra_channels=None, wait=None):
    event = Event(event_t, data)
    pubsub.publish("shoebot", event)
    for channel_name in extra_channels or []:
        pubsub.publish(channel_name, event)
    if wait is not None:
        channel = pubsub.subscribe(wait)
        channel.listen(wait)



### NEWSFEED #########################################################################################
# Code for parsing newsfeeds. 
# It wraps the Universal Feedparser by Mark Pilgrim.

# Author: Tom De Smedt.
# Copyright (c) 2007 by Tom De Smedt.
# See LICENSE.txt for details.

import os

from feedparser import feedparser

from url import URLAccumulator
from html import strip_tags
from cache import Cache

def clear_cache():
    Cache("newsfeed").clear()

### FAVORITE NEWSFEED ################################################################################

favorites = {}
try:
    path = os.path.join(os.path.dirname(__file__), "newsfeed.txt")
    for f in open(path).readlines():
        f = f.split(",")
        favorites[f[0].strip()] = f[1].strip()
except:
    pass

def favorite_url(name):

    if favorites.has_key(name):
        return favorites[name]

    for key in favorites:
        if key.lower().find(name.lower()) >= 0:
            return favorites[key]

    return None

favorite = favorite_url

### NEWSFEED #########################################################################################

class Newsfeed:
    
    """ Wrapper for the feedparser.FeedParserDict class.
    
    Ensures that Newsfeed.items redirects to Newsfeed.entries,
    and returns an empty string (by default) instead of raising an error
    when a key could not be found - this way we don't have to check 
    if a key exists before fetching its value.
    
    """
    
    def __init__(self, feed, none=u""):
        self._feed = feed
        self._none = none

    def __call__(self, *args):
        raise TypeError, "Newsfeed object not callable"

    def __repr__(self):
        return strip_tags(self._feed.__repr__())
    
    def __unicode__(self):
        return strip_tags(self._feed)
    
    def __str__(self):
        try: 
            s = self._feed.encode("utf-8")
        except:
            s = self._feed.__str__()
        return strip_tags(s)

    def __getitem__(self, item):
        try: return Newsfeed(self._feed.__getitem__(item))
        except:
            return Newsfeed(self._none)
    
    def has_key(self, key):
        return self._feed.has_key(key)
    
    def __iter__(self):
        return self._feed.__iter__()
    
    def __getattr__(self, a):
        if a == "items": a = "entries"
        try:
            a = self._feed.__getattr__(a)
            if isinstance(a, list): 
                a = [Newsfeed(x, self._none) for x in a]
            return Newsfeed(a)    
        except:
            return Newsfeed(self._none)

### NEWSFEED DOWNLOAD ################################################################################

class NewsfeedDownload(Newsfeed, URLAccumulator):
    
    """ Asynchronous cached Newsfeed.
    """
    
    def __init__(self, url, wait=10, asynchronous=False, cached=True, none=""):
        
        self._feed = None
        self._none = none
        
        if cached: 
            cache = "newsfeed"
        else:
            cache = None
            
        # Refresh cached news results every day.
        if cached and Cache(cache).age(url) > 0:
            Cache(cache).remove(url)
            
        URLAccumulator.__init__(self, url, wait, asynchronous, cache, ".xml")

    def load(self, data):
        
        parsed = feedparser.parse(data)
        Newsfeed.__init__(self, parsed, self._none)
            
def parse(url, wait=10, asynchronous=False, cached=True, none=""):
    
    nf = NewsfeedDownload(url, wait, asynchronous, cached, none)
    try:
        # Some shortcuts:
        nf.title = nf.channel.title
        nf.description = nf.channel.description
        nf.link = nf.channel.link
        nf.date = nf.channel.data
    except:
        pass
    
    return nf

"""
url = favorite_url("white house")
newsfeed = parse(url)

print "Channel:", newsfeed.channel.title
print "Channel description:", newsfeed.channel.description
print "Channel link:", newsfeed.channel.link
print "Channel date:", newsfeed.channel.date
print "Encoding:", newsfeed.encoding

for item in newsfeed.items:
    
    print "Title:", item.title
    print "Link:", item.link
    print "Description", item.description
    
    print "Date:", item.date
    print ">>>", item.date_parsed
    
    print "Author:", item.author
    print ">>", item.author_detail.name
    print ">>", item.author_detail.email
"""


#!/usr/bin/env python2

# This file is part of Shoebot.
# Copyright (C) 2009 the Shoebot authors
# See the COPYING file for the full license text.
#
#   Redistribution and use in source and binary forms, with or without
#   modification, are permitted provided that the following conditions are met:
#
#   Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
#   The name of the author may not be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
#   WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
#   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
#   EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
#   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
#   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
#   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
#   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
#   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''Assorted utility functions, mainly for color and font handling'''

from __future__ import division

'''
def rgb_to_hsl(r, g, b):
    """Converts RGB values to the HSL colourspace. """

    # Taken from Inkscape.
    # Copyright (C) 2006 Jos Hirth, kaioa.com
    # Subject to the terms of the GPLv2 or any later version.

    rgb_max = max (max (r, g), b)
    rgb_min = min (min (r, g), b)
    delta = rgb_max - rgb_min
    hsl = [0.0, 0.0, 0.0]
    hsl[2] = (rgb_max + rgb_min)/2.0
    if delta == 0:
        hsl[0] = 0.0
        hsl[1] = 0.0
    else:
        if hsl[2] <= 0.5:
            hsl[1] = delta / (rgb_max + rgb_min)
        else:
            hsl[1] = delta / (2 - rgb_max - rgb_min)
        if r == rgb_max:
            hsl[0] = (g - b) / delta
        else:
            if g == rgb_max:
                hsl[0] = 2.0 + (b - r) / delta
            else:
                if b == rgb_max:
                    hsl[0] = 4.0 + (r - g) / delta
        hsl[0] = hsl[0] / 6.0
        if hsl[0] < 0:
            hsl[0] = hsl[0] + 1
        if hsl[0] > 1:
            hsl[0] = hsl[0] - 1
    return hsl

def hue_2_rgb (v1, v2, h):
    """Helper function for converting HSL to RGB """

    # Taken from Inkscape.
    # Copyright (C) 2006 Jos Hirth, kaioa.com
    # Subject to the terms of the GPLv2 or any later version.

    if h < 0:
        h += 6.0
    if h > 6:
        h -= 6.0
    if h < 1:
        return v1 + (v2 - v1) * h
    if h < 3:
        return v2
    if h < 4:
        return v1 + (v2 - v1) * (4 - h)
    return v1

def hsl_to_rgb (h, s, l):
    """Converts HSL values to RGB."""

    # Taken from Inkscape.
    # Copyright (C) 2006 Jos Hirth, kaioa.com
    # Subject to the terms of the GPLv2 or any later version.

    rgb = [0, 0, 0]
    if s == 0:
        rgb[0] = l
        rgb[1] = l
        rgb[2] = l
    else:
        if l < 0.5:
            v2 = l * (1 + s)
        else:
            v2 = l + s - l*s
        v1 = 2*l - v2
        rgb[0] = hue_2_rgb (v1, v2, h*6 + 2.0)
        rgb[1] = hue_2_rgb (v1, v2, h*6)
        rgb[2] = hue_2_rgb (v1, v2, h*6 - 2.0)
    return rgb
'''

'''
def rgb_to_hsv(r, g, b):
    
    h = s = 0
    v = max(r, g, b)
    d = v - min(r, g, b)

    if v != 0:
        s = d / float(v)

    if s != 0:
        if   r == v : h = 0 + (g-b) / d
        elif g == v : h = 2 + (b-r) / d
        else        : h = 4 + (r-g) / d

    h = h * (60.0/360)
    if h < 0: 
        h = h + 1.0
    hsv=[h,s,v]    
    return hsv
'''

"""def hsv_to_rgb(h, s, v):
    
    ''' Hue, saturation, brightness to red, green, blue.
    http://www.koders.com/python/fidB2FE963F658FE74D9BF74EB93EFD44DCAE45E10E.aspx
    Results will differ from the way NSColor converts color spaces.
    '''
    
    if s == 0: return v, v, v
        
    h = h / (60.0/360)
    i =  floor(h)
    f = h - i
    p = v * (1-s)
    q = v * (1-s * f)
    t = v * (1-s * (1-f))
    
    if   i == 0 : r = v; g = t; b = p
    elif i == 1 : r = q; g = v; b = p
    elif i == 2 : r = p; g = v; b = t
    elif i == 3 : r = p; g = q; b = v
    elif i == 4 : r = t; g = p; b = v
    else        : r = v; g = p; b = q
    
    return list(r, g, b)
"""





#def rgba_to_argb(stringImage):
    #tempBuffer = [None]*len(stringImage) # Create an empty array of the same size as stringImage
    #tempBuffer[0::4] = stringImage[3::4]
    #tempBuffer[1::4] = stringImage[0::4]
    #tempBuffer[2::4] = stringImage[1::4]
    #tempBuffer[3::4] = stringImage[2::4]
    #stringImage = ''.join(tempBuffer)
    #return stringImage


class flushfile(object):
    '''
    Wrapper for file that flushes - used to flush stdout

    http://stackoverflow.com/questions/230751/how-to-flush-output-of-python-print
    '''
    def __init__(self, fd):
        self.fd = fd

    def write(self, x):
        ret=self.fd.write(x)
        self.fd.flush()
        return ret

    def writelines(self, lines):
        ret=self.writelines(line)
        self.fd.flush()
        return ret

    def flush(self):
        return self.fd.flush

    def close(self):
        return self.fd.close()

    def fileno(self):
        return self.fd.fileno()




# Cornu - last updated for NodeBox 1.9.3
# Cornu curves from Raph Levien.
# Cornu-to-Bezier code by Mark Meyer.
# The code is from http://casper.ghostscript.com/~raph/
# You can download Raph's curve editor from http://www.levien.com/spiro/.

from math import *

### CORNU ############################################################################################

# fit arc to pts (0, 0), (x, y), and (1, 0), return th tangent to
# arc at (x, y)
def fit_arc(x, y):
    th = atan2(y - 2 * x * y, y * y + x - x * x)
    return th

# find thetas tangent to path using local cspline logic
def local_ths(path, closed):
    if closed:
        result = []
    else:
        result = [0]
    for i in range(1 - closed, len(path) - 1 + closed):
        x0, y0 = path[(i + len(path) - 1) % len(path)]
        x1, y1 = path[i]
        x2, y2 = path[(i + 1) % len(path)]
        dx = x2 - x0
        dy = y2 - y0
        ir2 = dx * dx + dy * dy
        x = ((x1 - x0) * dx + (y1 - y0) * dy) / ir2
        y = ((y1 - y0) * dx - (x1 - x0) * dy) / ir2
        th = fit_arc(x, y) + atan2(dy, dx)
        result.append(th)
    if not closed:
        result.append(0)
    boundary_ths(path, result, closed)
    return result

# set the endpoint thetas so endpoint curves are circular arcs
def boundary_ths(path, ths, closed):
    if not closed:
        first_th = 2 * atan2(path[1][1] - path[0][1], path[1][0] - path[0][0]) - ths[1]
        ths[0] = first_th
        last_th = 2 * atan2(path[-1][1] - path[-2][1], path[-1][0] - path[-2][0]) - ths[-2]
        ths[-1] = last_th
   
# implementation adapted from cephes
def polevl(x, coef):
    ans = coef[-1]
    for i in range(len(coef) - 2, -1, -1):
        ans = ans * x + coef[i]
    return ans

sn = [
-2.99181919401019853726E3,
 7.08840045257738576863E5,
-6.29741486205862506537E7,
 2.54890880573376359104E9,
-4.42979518059697779103E10,
 3.18016297876567817986E11
]
sn.reverse()
sd = [
 1.00000000000000000000E0,
 2.81376268889994315696E2,
 4.55847810806532581675E4,
 5.17343888770096400730E6,
 4.19320245898111231129E8,
 2.24411795645340920940E10,
 6.07366389490084639049E11
]
sd.reverse()
cn = [
-4.98843114573573548651E-8,
 9.50428062829859605134E-6,
-6.45191435683965050962E-4,
 1.88843319396703850064E-2,
-2.05525900955013891793E-1,
 9.99999999999999998822E-1
]
cn.reverse()
cd = [
 3.99982968972495980367E-12,
 9.15439215774657478799E-10,
 1.25001862479598821474E-7,
 1.22262789024179030997E-5,
 8.68029542941784300606E-4,
 4.12142090722199792936E-2,
 1.00000000000000000118E0
]
cd.reverse()

fn = [
  4.21543555043677546506E-1,
  1.43407919780758885261E-1,
  1.15220955073585758835E-2,
  3.45017939782574027900E-4,
  4.63613749287867322088E-6,
  3.05568983790257605827E-8,
  1.02304514164907233465E-10,
  1.72010743268161828879E-13,
  1.34283276233062758925E-16,
  3.76329711269987889006E-20
]
fn.reverse()
fd = [
  1.00000000000000000000E0,
  7.51586398353378947175E-1,
  1.16888925859191382142E-1,
  6.44051526508858611005E-3,
  1.55934409164153020873E-4,
  1.84627567348930545870E-6,
  1.12699224763999035261E-8,
  3.60140029589371370404E-11,
  5.88754533621578410010E-14,
  4.52001434074129701496E-17,
  1.25443237090011264384E-20
]
fd.reverse()
gn = [
  5.04442073643383265887E-1,
  1.97102833525523411709E-1,
  1.87648584092575249293E-2,
  6.84079380915393090172E-4,
  1.15138826111884280931E-5,
  9.82852443688422223854E-8,
  4.45344415861750144738E-10,
  1.08268041139020870318E-12,
  1.37555460633261799868E-15,
  8.36354435630677421531E-19,
  1.86958710162783235106E-22
]
gn.reverse()
gd = [
  1.00000000000000000000E0,
  1.47495759925128324529E0,
  3.37748989120019970451E-1,
  2.53603741420338795122E-2,
  8.14679107184306179049E-4,
  1.27545075667729118702E-5,
  1.04314589657571990585E-7,
  4.60680728146520428211E-10,
  1.10273215066240270757E-12,
  1.38796531259578871258E-15,
  8.39158816283118707363E-19,
  1.86958710162783236342E-22
]
gd.reverse()


def fresnel(xxa):
    x = abs(xxa)
    x2 = x * x
    if x2 < 2.5625:
        t = x2 * x2
        ss = x * x2 * polevl(t, sn) / polevl(t, sd)
        cc = x * polevl(t, cn) / polevl(t, cd)
    elif x > 36974.0:
        ss = 0.5
        cc = 0.5
    else:
        t = pi * x2
        u = 1.0 / (t * t)
        t = 1.0 / t
        f = 1.0 - u * polevl(u, fn) / polevl(u, fd)
        g = t * polevl(u, gn) / polevl(u, gd)
        t = pi * .5 * x2
        c = cos(t)
        s = sin(t)
        t = pi * x
        cc = 0.5 + (f * s - g * c) / t
        ss = 0.5 - (f * c + g * s) / t
    if xxa < 0:
        cc = -cc
        ss = -ss
    return ss, cc

def eval_cornu(t):
    spio2 = sqrt(pi * .5)
    s, c = fresnel(t / spio2)
    s *= spio2
    c *= spio2
    return s, c

def mod_2pi(th):
    u = th / (2 * pi)
    return 2 * pi * (u - floor(u + 0.5))

def fit_cornu_half(th0, th1):
    if th0 + th1 < 1e-6:
        epsilon = 1e-6
        th0 += epsilon
        th1 += epsilon
    n_iter = 0
    n_iter_max = 21
    est_tm = 0.29112 * (th1 + th0) / sqrt(th1 - th0)
    l = est_tm * .9
    r = est_tm * 2
    while 1:
        t_m = .5 * (l + r)
        dt = (th0 + th1) / (4 * t_m)
        t0 = t_m - dt
        t1 = t_m + dt
        # invariant: t1^2 - t0^2 = th0 + th1
        s0, c0 = eval_cornu(t0)
        s1, c1 = eval_cornu(t1)
        chord_th = atan2(s1 - s0, c1 - c0)
        n_iter += 1
        if n_iter == n_iter_max:
            break
        if mod_2pi(chord_th - t0 * t0 - th0) < 0:
            l = t_m
        else:
            r = t_m
    chordlen = hypot(s1 - s0, c1 - c0)
    k0 = t0 * chordlen
    k1 = t1 * chordlen
    return t0, t1, k0, k1

def fit_cornu(th0, th1):
    if th1 > th0:
        return fit_cornu_half(th0, th1)
    elif th0 > th1:
        t0, t1, k0, k1 = fit_cornu_half(th1, th0)
        # ?
        return (-t1, -t0, -k1, -k0)
    else:
        # it's a circle
        return None

def draw_tan(x, y, th):
    dx = .2 * cos(th)
    dy = .2 * sin(th)
    print_pt(x - dx, y - dy, 'moveto')
    print_pt(x + dx, y + dy, 'lineto')

def draw_cornu(path, ths, closed, flat=False):
    cmd = 'moveto'
    for i in range(len(path) - 1 + closed):
        x0, y0 = path[i]
        x1, y1 = path[(i + 1) % len(path)]
        th = atan2(y1 - y0, x1 - x0)
        th0 = mod_2pi(ths[i] - th)
        th1 = mod_2pi(th - ths[(i + 1) % len(path)])
        flip = -1
        th1 += 1e-6
        if th1 < th0:
            th0, th1 = th1, th0
            flip = 1
        t0, t1, k0, k1 = fit_cornu_half(th0, th1)
        if flip == 1:
            t0, t1 = t1, t0
            k0, k1 = k1, k0
        s0, c0 = eval_cornu(t0)
        s0 *= flip
        s1, c1 = eval_cornu(t1)
        s1 *= flip
        chord_th = atan2(s1 - s0, c1 - c0)
        chordlen = hypot(s1 - s0, c1 - c0)
        rot = th - chord_th
        scale = hypot(y1 - y0, x1 - x0) / chordlen
        cs = scale * cos(rot)
        ss = scale * sin(rot)
        if flat:
            cmd = draw_cornu_flat(x0, y0, t0, t1, s0, c0, flip, cs, ss, cmd)
        else:
            cmd = draw_cornu_bezier(x0, y0, t0, t1, s0, c0, flip, cs, ss, cmd, scale, rot)
    if closed:
        pass # print 'closepath'
    #print 'stroke'

def draw_cornu_flat(x0, y0, t0, t1, s0, c0, flip, cs, ss, cmd):
    
    """ Raph Levien's code draws fast LINETO segments.
    """
    
    for j in range(0, 100):
        t = j * .01
        s, c = eval_cornu(t0 + t * (t1 - t0))
        s *= flip
        s -= s0
        c -= c0
        #print '%', c, s
        x = c * cs - s * ss
        y = s * cs + c * ss
        print_pt(x0 + x, y0 + y, cmd)
        cmd = 'lineto'
    return cmd
    
def draw_cornu_bezier(x0, y0, t0, t1, s0, c0, flip, cs, ss, cmd, scale, rot):

    """ Mark Meyer's code draws elegant CURVETO segments.
    """

    s = None
    for j in range(0, 5):
        # travel along the function two points at a time (at time t and t2)
        # the first time through we'll need to get both points
        # after that we only need the second point because the old second point
        # becomes the new first point
        t = j * .2
        t2 = t+ .2
        
        curvetime = t0 + t * (t1 - t0)
        curvetime2 = t0 + t2 * (t1 - t0)
        Dt = (curvetime2 - curvetime) * scale
        
        if not s:
            # get first point
            # avoid calling this again: the next time though x,y will equal x3, y3
            s, c = eval_cornu(curvetime)
            s *= flip
            s -= s0
            c -= c0
            # calculate derivative of fresnel function at point to get tangent slope
            # just take the integrand of the fresnel function
            dx1 =  cos(pow(curvetime, 2) + (flip * rot))  
            dy1 =  flip * sin(pow(curvetime, 2) + (flip *rot))
            # x,y = first point on function
            x = ((c * cs - s * ss) +x0)
            y = ((s * cs + c * ss) + y0)

        #evaluate the fresnel further along the function to look ahead to the next point
        s2,c2 = eval_cornu(curvetime2) 
        s2 *= flip
        s2 -= s0
        c2 -= c0

        dx2 = cos(pow(curvetime2, 2) + (flip * rot)) 
        dy2 = flip * sin(pow(curvetime2, 2) + (flip * rot))
        # x3, y3 = second point on function
        x3 = ((c2 * cs - s2 * ss)+x0)
        y3 = ((s2 * cs + c2 * ss)+y0)

        # calculate control points
        x1 = (x + ((Dt/3.0) * dx1))
        y1 = (y + ((Dt/3.0) * dy1))   
        x2 = (x3 - ((Dt/3.0) * dx2))
        y2 = (y3 - ((Dt/3.0) * dy2))

        if cmd == 'moveto':
            print_pt(x, y, cmd)
            cmd = 'curveto'
        print_crv(x1, y1, x2, y2, x3, y3)
                    
        dx1, dy1 = dx2, dy2
        x,y = x3, y3
        
    return cmd
    
# update thetas based on cornu splines
def tweak_ths(path, ths, closed):
    dks = []
    for i in range(len(path) - 1 + closed):
        x0, y0 = path[i]
        x1, y1 = path[(i + 1) % len(path)]
        th = atan2(y1 - y0, x1 - x0)
        th0 = mod_2pi(ths[i] - th)
        th1 = mod_2pi(th - ths[(i + 1) % len(path)])
        flip = -1
        th1 += 1e-6
        if th1 < th0:
            th0, th1 = th1, th0
            flip = 1
        t0, t1, k0, k1 = fit_cornu_half(th0, th1)
        if flip == 1:
            t0, t1 = t1, t0
            k0, k1 = k1, k0
        s0, c0 = eval_cornu(t0)
        s1, c1 = eval_cornu(t1)
        chordlen = hypot(s1 - s0, c1 - c0)
        scale = 1 / hypot(y1 - y0, x1 - x0)
        k0 *= scale
        k1 *= scale
        if i > 0:
            dk = k0 - last_k1
            dks.append(dk)
        else:
            first_k0 = k0
        last_k1 = k1
    if closed:
        dks.append(first_k0 - last_k1)
    for i in range(len(dks)):
        x0, y0 = path[i]
        x1, y1 = path[(i + 1) % len(path)]
        x2, y2 = path[(i + 2) % len(path)]
        chord1 = hypot(x1 - x0, y1 - y0)
        chord2 = hypot(x2 - x1, y2 - y1)
        ths[(i + 1) % len(path)] -= .5 * (dks[i] / (chord1 + chord2))

def csinterp(t0, t1, u):
    if u == 0: return (0, 0)
    elif u == 1: return (1, 0)
    c = cos(u * pi * 0.5)
    s = sin(u * pi * 0.5)

    #sigmoid interpolant:
    tau = t0 * c * c + t1 * s * s

    #linear interpolant:
    #tau = t0 * (1 - u) + t1 * u + 1e-9

    f = sin(u * tau) / sin(tau)
    ph = (1 - u) * tau
    return (f * cos(ph), f * sin(ph))

def draw_cspline(path, ths):
    cmd = 'moveto'
    for i in range(len(path) - 1):
        x0, y0 = path[i]
        x1, y1 = path[i + 1]
        th = atan2(y1 - y0, x1 - x0)
        th0 = mod_2pi(ths[i] - th)
        th1 = mod_2pi(th - ths[i + 1])
        scale = hypot(y1 - y0, x1 - x0)
        rot = th
        cs = scale * cos(rot)
        ss = scale * sin(rot)
        for j in range(0, 100):
            t = j * .01
            c, s = csinterp(th0, th1, t)
            x = c * cs - s * ss
            y = s * cs + c * ss
            print_pt(x0 + x, y0 + y, cmd)
            cmd = 'lineto'

### NODEBOX BINDINGS #################################################################################

def print_pt(x, y, cmd):
    x = 100 * x
    y = 100 * y
    if (cmd == 'moveto'):        
        _ctx.moveto(x, y)
    elif (cmd == 'lineto'):
        _ctx.lineto(x, y)
    else:
        print cmd, x, y

def print_crv(x1, y1, x2, y2, x3, y3):
    x1 *= 100
    y1 *= 100
    x2 *= 100
    y2 *= 100
    x3 *= 100
    y3 *= 100
    _ctx.curveto(x1, y1, x2, y2, x3, y3)

def dot_pt(x, y):
    _ctx.oval(x-3,y-3, 6, 6) 

def relativise(path):

    for i in range(len(path)):
        x, y = path[i]
        x *= 0.01 * _ctx.WIDTH
        y *= 0.01 * _ctx.HEIGHT
        
        #Points on the path that have identical coordinates
        #generate a ZeroDivisionError
        for point in path:
            if (x,y) == point:
                x += 0.0000000001
                y += 0.0000000001
        
        path[i] = (x,y)
        
    return path
    
def path(points, close=False, tweaks=20, flat=False, draw=False, helper=False):

    # To make sure we don't change the original, make a copy of the 
    # given points
    points = list(points)

    points = relativise(points)
        
    ths = local_ths(points, close)
    for i in range(tweaks):
        boundary_ths(points, ths, close)
        tweak_ths(points, ths, close)
    boundary_ths(points, ths, close)

    _ctx.autoclosepath(False)
    _ctx.beginpath()
    draw_cornu(points, ths, close, flat)
    p =  _ctx.endpath(draw=draw)

    #beginpath()
    #draw_cspline(path, ths)
    #endpath()

    if helper:
        _ctx.beginpath()
        for i in range(len(points)):
            x, y = points[i]
            draw_tan(x, y, ths[i])
        _ctx.endpath()
        for x,y in points:
            dot_pt(x*100,y*100)
    
    return p
            
def drawpath(p, close=False, tweaks=20, points=False, flat=False):
    
    path(p, close, tweaks, flat, draw=True, helper=points)

### PSYCO SPECIALIZATION #############################################################################

try:
    import psyco
    for f in [
        fit_arc, local_ths, boundary_ths, polevl, fresnel, eval_cornu, 
        mod_2pi, fit_cornu_half, fit_cornu, draw_tan, draw_cornu, 
        draw_cornu_flat, draw_cornu_bezier,
        tweak_ths, csinterp, draw_cspline, print_pt, relativise
        ]:
        psyco.bind(f)
    #print "using psyco"
except:
    pass



tuio = ximport("tuio")

size(640, 480)
speed(30)

def setup():
    global tracking
    tracking = tuio.Tracking()

def draw():
    global tracking
    tracking.update()
    fontsize(10)
    for obj in tracking.objects():
        x = obj.xpos * WIDTH
        y = obj.ypos * HEIGHT
        rotate(obj.angle)
        rect(x, y, 20, 20)
        reset()
        text(obj, x, y)

def stop():
    global tracking
    tracking.stop()
